#!/usr/local/bin/php
<?php
/*!
 *  @file
 *
 *  @brief	setup HWS database for use
 *
 *  <p>
 *  <code>
 *  usage: setup_db.php \<DOCROOT\>
 *  </code>
 *  </p>
 *
 *  @returns
 *  Exit code 0 indicates the HWS database has been created; code 1 is returned
 *  in case an error occurred while creating the database.
 */

//-------------------------------- functions -----------------------------------

function usage()
{
	global $my_basename;

	fprintf(STDOUT, "
usage: %s <DOCROOT>

Setup the database for the HWS system. This operation involves creating the
HWS database on the hosting machine, creating the necessary tables, and loading
certain tables with initial data.
The document root of HWS is given by DOCROOT. The settings within this document
root are used for accessing the machine that will be hosting the HWS database.
", $my_basename);

	exit(1);
}

// create database, tables, load initial data, etc.
function setup_database($db, $dbname, $tables, $init_data, $triggers)
{
	global $my_basename;

	fprintf(STDOUT, "Creating HWS database...\n");

	if (!$db) return (false);

	if ($db->Connect($dbname)) {
		fprintf(STDERR, "%s: %s (%s)\n",
				$my_basename,
				E_DBEXIST,
				$dbname);
		return (false);
	}

	// connect to DBMS, nothing more
	if (!$db->Connect('')) {
		fprintf(STDERR, "%s: %s (%s)\n",
				$my_basename,
				E_DBCONNECT,
				$db::DB_IMPL);
		return (false);
	}

	$ret = $db->CreateDatabase($dbname);
	if ($ret) {
		fprintf(STDOUT, "  -> %s\n", $dbname);

		$ret = $db->Connect($dbname);
		if ($ret) {
			$ret = create_tables($db, $tables);
			if ($ret) {
				$ret = insert_data($db, $init_data);
			}
			if ($ret) {
				$ret = create_triggers($db, $triggers);
			}
		} else {
			fprintf(STDERR, "%s: %s\n", $my_basename, E_DBSELECT);
		}
	} else {
		fprintf(STDERR, "%s: %s (%s)\n",
				$my_basename,
				E_DBCREATE,
				$dbname);
	}
	return $ret;
}

// create table definitions
function create_tables($db, $tables)
{
	global $my_basename;
	$ret = true;

	fprintf(STDOUT, "Creating database tables...\n");

	if (!$db) return (false);

	if (!empty($tables)) {
		foreach ($tables as $table => $cols) {
			$ret = $db->CreateTable($table, $cols);
			if ($ret) {
				fprintf(STDOUT, "  -> %s\n", $table);
			} else {
				fprintf(STDERR, "%s: %s (%s)\n",
					$my_basename, E_TBLCREATE, $table);
				$ret = false;	// flag error
				break;
			}
		}
	} else {
		fprintf(STDOUT, "no table definitions\n");
		// not considered an error
	}
	return $ret;
}

// create database triggers
function create_triggers($db, $triggers)
{
	global $my_basename;
	$ret = true;

	fprintf(STDOUT, "Creating database triggers...\n");

	if (!$db) return (false);

	if (!empty($triggers)) {
		foreach ($triggers as $tname => $tdefs) {
			$ret = $db->CreateTrigger($tname,
						  $tdefs['table'],
						  $tdefs['event'],
						  $tdefs['when'],
						  $tdefs['stmt']);
			if ($ret) {
				fprintf(STDOUT, "  -> %s\n", $tname);
			} else {
				fprintf(STDERR, "%s: %s (%s)\n",
					$my_basename, E_TRIGCREATE, $tname);
				$ret = false;	// flag error
				break;
			}
		}
	} else {
		fprintf(STDOUT, "no trigger definitions\n");
		// not considered an error
	}
	return $ret;
}

// load initial table data
function insert_data($db, $tables)
{
	global $my_basename;
	$ret = true;

	fprintf(STDOUT, "Creating initial table data...\n");

	if (!$db) return (false);

	if (!empty($tables)) {
		foreach ($tables as $table => $rows) {
			fprintf(STDOUT, "  -> %s\n", $table);

			foreach ($rows as $row) {
				/*
				 * Each row is a string of column/value pairs
				 * separated by a semicolon. We need to convert
				 * this string to an (associative) array, so we
				 * can feed it to the database Insert() func.
				 */
				$fields = string_to_array($row);
				$ret = $db->Insert($table, $fields);
				if ($ret !== false) {
					fprintf(STDOUT, "     %s\n", $row);
				} else {
					fprintf(STDERR, "%s: %s (%s)\n",
						$my_basename,
						E_DBINSERT,
						$row);
					$ret = false;	// flag error
					break;
				}
			}
		}
	} else {
		fprintf(STDOUT, "no initial table data\n");
		// not considered an error
	}
	return $ret;
}

// convert string of column definitions to associative array
function string_to_array($str)
{
	/*
	 * The given string is assumed to have the following format:
	 *
	 *   <column> "=" <value> (";" <column> "=" <value>)*
	 *
	 * Each <column> (name) must be unique; <value> is an arbitrary string
	 * but should not contain the semicolon separator.
	 * The resulting array will hold key/value pairs where the extracted
	 * <column>s are the keys and the associated <value> its value. Leading
	 * and trailing whitespace is removed on all <value>s.
	 */
	$tmp   = array();	// auxillary array
	$cols  = array();
	$pairs = array_filter(explode(';', $str));

	foreach ($pairs as $p) {
		$tmp = explode('=', $p);
		$key = trim($tmp[0]);
		$val = trim($tmp[1]);
		$cols[$key] = $val;
	}
	return $cols;
}

// add first news item
function headlines($db)
{
	$datetime = explode(' ', date("Y-m-d H:i:s", time()));
	$subject  = 'HWS Installation';
	$message  = 'Congratulations, HWS has been installed on your system.
		     <br><br>
		     Have Fun!';

	$db->Insert('news', array('date'    => $datetime[0],
				  'time'    => $datetime[1],
				  'subject' => $subject,
				  'message' => $message));
}

//----------------------------- start of script --------------------------------

$my_basename = basename($argv[0]);
$exe_path    = dirname(realpath($argv[0]));
$docroot     = '';

if ($argc < 2 || $argc > 4) usage();

if ($argc == 2) $docroot = $argv[1];

// work from within document root:
if (is_dir($docroot)) {
	chdir($docroot) || exit(1);
} else {
	fprintf(STDERR, "%s: %s: directory not found\n", $my_basename,$docroot);
	exit(1);
}

require 'init.php';

$dbms = trim(DBMS);
if (empty($dbms)) {
	fprintf(STDERR, "%s: no DBMS set\n", $my_basename);
	exit(1);
}

include $exe_path . '/hws_tables.'   . $dbms . '.inc';
include $exe_path . '/hws_triggers.' . $dbms . '.inc';
include $exe_path . '/init_tables.inc';


fprintf(STDOUT, "\nHWS Database Setup");
fprintf(STDOUT, "\n------------------\n");

$ret = setup_database($DB, DB_NAME, $HWS_TABLES, $INIT_DATA, $HWS_TRIGGERS);
if ($ret) {
	headlines($DB);
}

exit($ret == false ? 1 : 0)	// convert to UNIX exit code
?>
