<?php
/*!
 *  @file
 *
 *  @brief	read devices from uploaded file
 *
 *  @date	November 7, 2015
 */

require 'inc/func_dtd.php';
require 'inc/PDD.inc';

has_permissions('w') || fatal(E_NOPERM);

$gid = validate_integer($_POST['gid']);
$tid = validate_integer($_POST['tid']);

$dtd = get_dtd_pathname($tid);
require $dtd;

$infile = $_FILES['import_file']['tmp_name'];

echo '
	<h4>Importing devices for '.filter_string($DB->GetDeviceTypeDesc($tid)).'...</h4>
';

/*
 * First some basic sanity checks.
 * Abort script on error.
 */
if (file_exists($infile)) {
	if (filesize($infile) > IMPORT_FILE_MAX) {
		fatal(E_FTOOLARGE);
	}
} else {
	fatal(E_NOFILE);
}

/*
 * Pre-process CSV file. Read file into array (one line
 * per entry), trim leading and trailing whitespace, and
 * eliminate blank lines.
 */
$csv = array();		// CSV entries
$fp = fopen($infile, 'r');
if ($fp) {
	while (($s = fgets($fp)) !== false) {
		$s = trim($s);
		if (!empty($s)) {
			array_push($csv, $s);
		}
	}
	fclose($fp);
} else {
	fatal(E_FOPEN);
}

/*
 * More sanity checks. Check whether there's an entry
 * in the file specifying column names. This must be
 * the first entry in the file. In addition, the
 * column names must match for the device type being
 * processed. Again, the script is aborted on error.
 */
$csv_col_names = str_getcsv($csv[0]);
$num_cols = count($csv_col_names);
if ($num_cols > 0) {
	/*
	 * Determine column names of physical device and the device type
	 * (i.e., serial number, part number, model etc).
	 */
	$all_col_names =
		array_values(			// -> eliminate array keys
			array_map(		// -> collect column names
				function($e)	// -> get one column name
				{
					return $e[1];
				},
				array_merge($PDD['features'], $DTD['features'])
			)
		);
	/*
	 * Now we know which column names are possible, we can
	 * check them against the ones used in the CSV file.
	 */
	$num_intersect = count(array_intersect($all_col_names, $csv_col_names));
	if ($num_intersect != $num_cols) {
		fatal(E_CSVHDR);
	}
} else {
	fatal(E_CSVNOHDR);
}

/*
 * Begin processing CSV entries.
 */
unset($csv[0]);		// skip column names entry
$num_total  = count($csv);
$num_insert = 0;
$num_skip   = 0;
$num_fail   = 0;
foreach ($csv as $entry) {
	$named_entry = array_combine($csv_col_names, str_getcsv($entry));

	// get physical device fields & values:
	$physical_feats = extract_features($PDD['features'], $named_entry);
	$did = $DB->GetPhysicalDeviceId($gid, $tid,
				$physical_feats['serial_number']['value']);
	if ($did != 0) {
		++$num_skip;
		continue;	// already exists
	}

	$model_feats = extract_features($DTD['features'], $named_entry);
	$msids = resolve_feature_ids($DTD['table'], $model_feats);
	foreach ($DTD['features'] as $k => $v) {
		// $v[1]: feature column name
		// $v[3]: feature data type
		// $v[4]: type-specific data (e.g., mapping table)
		// $v[6]: default unit
		// $v[7]: unit mapping table

		$ftvalue = trim($msids["$v[1]"]);
		if (empty($ftvalue)) continue;	// skip empty values

		if ($v[3] == TEXT && !empty($v[7])) {
			// resolve unit
			$num = convert_value($ftvalue, $v[6], $v[7]);
			if ($value === false) {
				continue 2;	// faulty, skip CSV entry
			} else {
				$msids["$v[1]"] = $num;
			}
		} else if ($v[3] == RBTN && !empty($v[4])) {
			$mapping = array('label' => $ftvalue);
			$value = $DB->FetchByValue($v[4], $mapping);
			$msids["$v[1]"] = $value[0]['value'];
		} else if ($v[3] == CBOX) {
			unset($msids["$v[1]"]);
			// turn comma-separated list of checbox options
			// into an array; needed when resolving seq. numbers
			$opts = array();
			if (!empty($model_feats[$k]['value'])) {
				$opts = explode(',', $model_feats[$k]['value']);
			}
			$model_feats[$k]['value'] = $opts;
		}
	}
	$jtsn = resolve_jtable_sqn($DTD['table'], $model_feats);
	$msids = array_merge($msids, $jtsn);

	$row = $DB->FetchByValue($DTD['table'], $msids);
	if (empty($row)) {
		$msid = $DB->Insert($DTD['table'], $msids);
	} else {
		$msid = $row[0]['id'];
	}

	$phids = resolve_feature_ids('physical', $physical_feats);

	$fields = array('hwgroup_id'     => $gid,
			'device_type_id' => $tid,
			'model_spec_id'  => $msid,
			'status_id'      => $phids['status_id'],
			'location_id'    => $phids['location_id'],
			'serial_number'  => $physical_feats['serial_number']['value'],
			'barcode'        => $physical_feats['barcode']['value']);
	$ret = $DB->Insert('physical', $fields);
	$ret !== false ? ++$num_insert : ++$num_fail;
}

echo '
	<form name="model_import" method="post" action="?component=model&action=list&gid='.$gid.'&tid='.$tid.'">
		<table class="std_input">
			<tr>
				<td class="std_input_title">Total Number of Records:</td>
				<td class="std_input_item">'.$num_total.'</td>
			</tr>
			<tr>
				<td class="std_input_title">Records inserted:</td>
				<td class="std_input_item">'.$num_insert.'</td>
			</tr>
			<tr>
				<td class="std_input_title">Records skipped:</td>
				<td class="std_input_item">'.$num_skip.'</td>
			</tr>
			<tr>
				<td class="std_input_title">Failed insertions:</td>
				<td class="std_input_item">'.$num_fail.'</td>
			</tr>
			<tr>
				<td class="std_input_buttons" colspan="2"><input type="submit" value="OK"></td>
			</tr>
		</table>
	</form>
';
?>
