<?php
/*!
 *  @file
 *
 *  @brief	implementation of MySQL database access
 *
 *  @date	July 4, 2015
 */

/*!
 *  Implementation of MySQL database access for HWS. This class defines
 *  functions, properties, and queries which are specific for the MySQL
 *  database implementation.
 */
class MySqlDatabase extends SqlDatabase
{
	function __construct($host, $user, $pass)
	{
		parent::__construct($host, $user, $pass);
	}

	function __destruct()
	{
		$this->Close();

		parent::__destruct();
	}

	public function GetDbInfo()
	{
		$db_info = array();

		$db_info['implementation'] = self::DB_IMPL;
		$db_info['host'] = $this->mHost;

		if ($this->mLink) {
			$db_info['state'] = true;
			$db_info['version'] = 
					@mysql_get_server_info($this->mLink);
		} else {
			$db_info['state'] = false;
			$db_info['version'] = '-';
		}

		$db_info['uptime'] = $this->GetUptime();
		if (empty($db_info['uptime'])) {
			$db_info['uptime'] = 'N/A';
		}

		return $db_info;
	}

	/*!
	 *  @todo
	 *  Only supports one trigger statement. Multi-statements only
	 *  supported when PHP was build with native MySQL driver (mysqlnd).
	 */
	function CreateTrigger($trigger, $table, $event, $when, $stmts)
	{
		if (trim($trigger) == false ||
		    trim($table)   == false)
		{
			return false;
		}

		$ev = strtoupper(trim($event));
		if (strcmp($ev, 'INSERT') != 0 &&
		    strcmp($ev, 'UPDATE') != 0 &&
		    strcmp($ev, 'DELETE') != 0)
		{
			return false;
		}

		$wh = strtoupper(trim($when));
		if (strcmp($wh, 'BEFORE') != 0 &&
		    strcmp($wh, 'AFTER')  != 0)
		{
			return false;
		}

		// format query:
		$sql  = 'CREATE TRIGGER ' . $trigger . ' ' . $wh . ' ' . $ev;
		$sql .= ' ON ' . $table . ' FOR EACH ROW '. implode(' ',$stmts);
		return $this->Query($sql);
	}

	public function Connect($db)
	{
		$ret = false;

		$this->mLink = @mysql_connect($this->mHost,
						$this->mUser,
						$this->mPass);
		if ($this->mLink) {
			if (trim($db)) {
				$ret = @mysql_select_db($db, $this->mLink);
				if ($ret) {
					$this->mDatabase = trim($db);
				}
			} else {
				$ret = true;
				// no error if database was not set,
				// then we're only connected to DBMS
			}
		}
		return ($ret);
	}

	public function Close()
	{
		if ($this->mLink) {
			@mysql_close($this->mLink);
		}
	}

	public function Insert($table, $fields)
	{
		$fld_set = array();
		$id = 0;

		if (!$this->mLink) {
			return false;
		}

		// format query:
		$sql = 'INSERT INTO ' . $table . ' SET ';
		foreach ($fields as $k => $val) {
			$parm = $this->FmtSqlParameter($table, $k, $val);
			if ($parm === false) {
				fatal($this->FmtLastError('Insert',
							print_r($parm)));
			}

			array_push($fld_set, $k.'='.$parm);
		}
		$sql .= implode(',', $fld_set);
		
		$ret = $this->Query($sql);
		if ($ret) {
			$id = @mysql_insert_id($this->mLink);
		} else {
			fatal($this->FmtLastError('Insert', $sql));
		}

		return ($id >= 0 ? $id : false);
	}

	public function Update($table, $id, $fields)
	{
		$fld_set = array();

		if (!$this->mLink) {
			return false;
		}

		$sql = 'UPDATE ' . $table . ' SET ';

		// format query:
		foreach ($fields as $k => $val) {
			$parm = $this->FmtSqlParameter($table, $k, $val);
			if ($parm === false) {
				fatal($this->FmtLastError('Update',
							print_r($parm)));
			}
			array_push($fld_set, $k.'='.$parm);
		}
		$sql .= implode(',', $fld_set);

		$sql .= ' WHERE id=' . $id; // row to act upon

		$ret = $this->Query($sql);
		if ($ret) {
			return true;
		} else {
			fatal($this->FmtLastError('Update', $sql));
		}
	}

	public function EscapeString($str)
	{
		return @mysql_real_escape_string($str, $this->mLink);
	}

	public function FmtSqlParameter($table, $column, $value)
	{
		if (is_null($value))    return 'NULL';
		// check null first as it's not considered scalar

		if (!is_scalar($value)) return false;

		$info = $this->GetInfoSchema($table);
		switch ($info[$column]['type']) {
		case 'bool':
		case 'int':
			return (int)$value;
		case 'float':
			return (float)$value;
		case 'string':
			return '\''.$this->EscapeString($value).'\'';
		default:
			return $value;
		}
	}

	public function FmtLastError($context, $data)
	{
		$err  = '<h4>Error</h4>';
		$err .= @mysql_error($this->mLink);
		$err .= ' ('.@mysql_errno($this->mLink).')';
		$err .= '<h4>Database</h4>';
		$err .= $this->mDatabase;
		$err .= '<h4>Backend</h4>';
		$err .= self::DB_IMPL;
		$err .= '<h4>Context</h4>';
		$err .= empty($context) ? 'N/A' : $context;
		$err .= '<h4>Context Data</h4>';
		$err .= empty($data) ? 'N/A' : filter_string($data);

		return $err;
	}

	public function GetInfoSchema($table)
	{
		$info = $this->GetInfoSchemaRaw($table);

		foreach ($info as &$col_info) {
			switch (strtolower($col_info['type'])) {
			case 'boolean':
			case 'bool':
				$col_info['type'] = 'bool'; break;
			case 'int':
			case 'integer':
			case 'tinyint':
			case 'smallint':
			case 'mediumint':
			case 'bigint':
				$col_info['type'] = 'int'; break;
			case 'float':
			case 'double':
			case 'real':
			case 'decimal':
			case 'numeric':
				$col_info['type'] = 'float'; break;
			default:
				$col_info['type'] = 'string'; break;
			}
		}
		unset($col_info);

		return $info;
	}

	/*!
	 *  \name Physical Devices
	 *  @{
	 */

	public function GetNumberOfPhysicalDevices($gid, $tid, $msid)
	{
		// format device filter statements, if necessary:
		$filters = array();
		if (!empty($gid)) {
			array_push($filters, 'hwgroup_id='.$gid);
		}
		if (!empty($tid)) {
			array_push($filters, 'device_type_id='.$tid);
		}
		if (!empty($msid)) {
			array_push($filters, 'model_spec_id='.$msid);
		}

		// query to filter physical devices:
		$sql = 'SELECT SQL_CALC_FOUND_ROWS id FROM physical';
		if (count($filters) > 0) {
			$sql .= ' WHERE ' . implode(' AND ', $filters);
		}
		$ret = $this->Query($sql);

		$sql = 'SELECT FOUND_ROWS() count';
		$ret = $this->Query($sql);
		if ($ret) {
			return $ret[0]['count'];
		} else {
			return false;
		}
	}

	public function GetPhysicalDeviceId($gid, $tid, $sn)
	{
		$sql  = 'SELECT SQL_CALC_FOUND_ROWS id FROM physical';
		$sql .= ' WHERE hwgroup_id=' . $gid;
		$sql .= '   AND device_type_id=' . $tid;
		$sql .= '   AND serial_number=\'' . $sn . '\'';
		$rows = $this->Query($sql);
	
		$sql = 'SELECT FOUND_ROWS() count';
		$ret = $this->Query($sql);
		if ($ret) {
			$nmatches = $ret[0]['count'];
			if ($nmatches > 1) {		// multiple matches
				$ret_val = -$nmatches;
			} else if ($nmatches == 1) {	// one match
				$ret_val = $rows[0]['id'];
			} else {			// no match
				$ret_val = 0;
			}
		} else {
			$ret_val = 0;
		}
		return ($ret_val);
	}

	/*!
	 *  end Physical Devices
	 *  @}
	 */

	protected function Query($sql)
	{
		$rows = array();

		if (!$this->mLink) {
			return false;
		}

		$result = @mysql_query($sql, $this->mLink);
		if ($result === false) {
			fatal($this->FmtLastError('Query', $sql));
		}

		if ($result === true) {
			return $result;
		} else {
			// retrieve results:
			while ($r = @mysql_fetch_assoc($result)) {
				array_push($rows, $r);
			}
			return $rows;
		}
	}

	/*!
	 *  Return data type of column. This is a helper function which
	 *  filters the data type length that may be part of the data type
	 *  specification. The length is assumed to be given in parenthesis.
	 *
	 *  @param[in] $str Column definition string.
	 *
	 *  @returns
	 *  Data type string.
	 */
	private function GetCharType($str)
	{
		$regex = '/\([0-9]+\)/';

		return preg_replace($regex, '//', $str);
	}

	/*!
	 *  Determine length of a @c CHAR type column. This is a helper
	 *  function which tries to extract the length part of a
	 *  character-based table column. Such a column has the substring
	 *  @c CHAR in its name and the maximum length in parenthesis.
	 *
	 *  @param[in] $str Column definition string.
	 *
	 *  @returns
	 *  Maximum length of @c CHAR column or 0 if undetermined.
	 */
	private function GetCharTypeLength($str)
	{
		$regex = '/^[A-Z]*CHAR[A-Z]*\(([0-9]+)\)$/i';

		if (preg_match($regex, $str, $match)) {
			return $match[1];
		} else {
			return 0;
		}
	}

	/*!
	 *  Retrieve untranslated information schema of table. This function
	 *  is the same as GetInfoSchema() except the table information is
	 *  not translated.
	 *
	 *  @sa
	 *  GetInfoSchema()
	 */
	private function GetInfoSchemaRaw($table)
	{
		$info = array();

		$sql = 'DESCRIBE ' . $table;
		$ret = $this->Query($sql);
		if ($ret) {
			foreach ($ret as $row) {
				$name = $row['Field'];
				$type = $this->GetCharType($row['Type']);
				$len  = $this->GetCharTypeLength($row['Type']);

				// prepare results in the proper format:
				$info[$name] = array(
						'type' => $type,
						'max_length' => $len,
						'default' => $row['Default']
					       );
			}
		}
		return $info;
	}

	/*!
	 *  Return uptime of database.
	 *
	 *  @returns
	 *  String expressing uptime in years, months, days, hours, minutes,
	 *  and seconds. If years, months, or days equals zero then it's
	 *  omitted from the output string. So, the output is like this:
	 *  <pre><code>
	 *  [\<years\>, ][\<months\>, ][\<days\>, ]\<hours\>:\<min\>:\<sec\>
	 *  </code></pre>
	 *
	 *  An empty string is returned if the uptime could not be determined.
	 */
	private function GetUptime()
	{
		$uptime = '';

		$sql = 'SHOW STATUS LIKE \'Uptime\'';
		$ret = $this->Query($sql);
		if ($ret) {
			// uptime of MySQL is returned in seconds
			$sec = $ret[0]['Value'];
			// convert seconds to year, months etc;
			// the following is kind of a hack from
			// the PHP manual.
			$d1 = new DateTime();
			$d2 = new DateTime();
			$d2->add(new DateInterval('PT'.$sec.'S'));
			$interval = $d2->diff($d1);

			$fmt = '';
			if ($interval->y != 0) $fmt .= '%y years, ';
			if ($interval->m != 0) $fmt .= '%m months, ';
			if ($interval->d != 0) $fmt .= '%d days, ';
			$fmt .= '%H:%I:%S';
			$uptime = $interval->format($fmt);
		}
		return $uptime;
	}

	const DB_IMPL = 'MySQL';
}
?>
