<?php
/*!
 *  @file
 *
 *  @brief	implementation of PostgreSQL database access
 *
 *  @date	July 4, 2015
 */

/*!
 *  Implementation of PostgreSQL database access for HWS. This class defines
 *  functions, properties, and queries which are specific for the PostgreSQL
 *  database implementation.
 */
class PgSqlDatabase extends SqlDatabase
{
	function __construct($host, $user, $pass)
	{
		parent::__construct($host, $user, $pass);
	}

	function __destruct()
	{
		if ($this->mLink) {
			@pg_close($this->mLink);
		}
		parent::__destruct();
	}

	public function GetDbInfo()
	{
		$db_info = array();

		$db_info['implementation'] = self::DB_IMPL;
		$db_info['host'] = $this->mHost;

		if ($this->mLink) {
			$db_info['state'] = true;
			$db_info['version'] =
					@pg_version($this->mLink)['client'];
		} else {
			$db_info['state'] = false;
			$db_info['version'] = '-';
		}

		$db_info['uptime'] = $this->GetUptime();
		if (empty($db_info['uptime'])) {
			$db_info['uptime'] = 'N/A';
		}

		return $db_info;
	}

	function CreateTrigger($trigger, $table, $event, $when, $stmts)
	{
		if (trim($trigger) == false ||
		    trim($table)   == false)
		{
			return false;
		}

		$ev = strtoupper(trim($event));
		if (strcmp($ev, 'INSERT') != 0 &&
		    strcmp($ev, 'UPDATE') != 0 &&
		    strcmp($ev, 'DELETE') != 0)
		{
			return false;
		}

		$wh = strtoupper(trim($when));
		if (strcmp($wh, 'BEFORE') != 0 &&
		    strcmp($wh, 'AFTER')  != 0)
		{
			return false;
		}

		// format query for trigger function:
		$func = 'func_'.$trigger;	// name of trigger function
		$sql  = 'CREATE FUNCTION '.$func.'() RETURNS trigger AS $$';
		$sql .= ' BEGIN ';
		$sql .= implode(';', $stmts);
		$sql .= '; END;';
		$sql .= ' $$ LANGUAGE plpgsql';
		$ret = $this->Query($sql);
		if ($ret) {
			// format query for trigger:
			$sql  = 'CREATE TRIGGER '.$trigger.' '.$wh.' '.$ev;
			$sql .= ' ON '.$table . ' FOR EACH ROW';
			$sql .= ' EXECUTE PROCEDURE '.$func.'()';
			$ret = $this->Query($sql);
		}
		return $ret;
	}

	public function Connect($db)
	{
		$ret = false;

		if (trim($db)) {
			$db_parm = 'dbname=' . trim($db);
		}

		$this->mLink = @pg_connect(
				'host=' . $this->mHost . ' ' .
				'user=' . $this->mUser . ' ' .
				'password=\'' . $this->mPass . '\' ' .
				$db_parm);
		if ($this->mLink) {
			$this->mDatabase = trim($db);
			$ret = true;
		}
		return ($ret);
	}

	public function Close()
	{
		if ($this->mLink) {
			@pg_close($this->mLink);
		}
	}

	public function Insert($table, $fields)
	{
		$fld_set = array();
		$id = 0;

		if (!$this->mLink) {
			return false;
		}

		// pg_insert() converts values in $fields array
		$ret = @pg_insert($this->mLink, $table, $fields);
		if ($ret) {
			$sql = "SELECT currval(pg_get_serial_sequence('$table','id'))";
			$rows = $this->Query($sql);
			$id = $rows[0]['currval'];
		} else {
			// format context data:
			$cdata = $table . ': ' . print_r($fields, true);
			fatal($this->FmtLastError('Insert', $cdata));
		}
		return ($id >= 0 ? $id : false);
	}

	public function Update($table, $id, $fields)
	{
		$fld_set = array();

		if (!$this->mLink) {
			return false;
		}

		// pg_update() converts values in $fields array
		$ret = @pg_update($this->mLink, $table, $fields,
					array('id' => $id));
		if (!$ret) {
			$cdata = $table . '['.$id.']: ' .print_r($fields,true);
			fatal($this->FmtLastError('Update', $cdata));
		}

		return true;
	}

	public function EscapeString($str)
	{
		return @pg_escape_string($this->mLink, $str);
	}

        public function FmtSqlParameter($table, $column, $value)
	{
		if (is_null($value)) return 'NULL';
		// check null first as it's not considered scalar

		if (!is_scalar($value)) return false;

		$info = $this->GetInfoSchema($table);
		switch ($info[$column]['type']) {
		case 'bool':
			// Postgres uses "TRUE" and "FALSE" for boolean:
			return ($value == true ? 'TRUE' : 'FALSE');
		case 'int':
			return (int)$value;
		case 'float':
			return (float)$value;
		case 'string':
			return '\''.$this->EscapeString($value).'\'';
		default:
			return $value;
		}
	}

	public function FmtLastError($context, $data)
	{
		$ec = @pg_last_error($this->mLink);

		$err  = '<h4>Error</h4>';
		$err .= empty($ec) ? 'None' : filter_string($ec);
		$err .= '<h4>Database</h4>';
		$err .= $this->mDatabase;
		$err .= '<h4>Backend</h4>';
		$err .= self::DB_IMPL;
		$err .= '<h4>Context</h4>';
		$err .= empty($context) ? 'N/A' : filter_string($context);
		$err .= '<h4>Context Data</h4>';
		$err .= empty($data) ? 'N/A' : filter_string($data);

		return $err;
	}

	public function GetInfoSchema($table)
	{
		$info = $this->GetInfoSchemaRaw($table);

		foreach ($info as &$col_info) {
			switch (strtolower($col_info['type'])) {
			case 'boolean':
			case 'bool':
				$col_info['type'] = 'bool'; break;
			case 'int':
			case 'integer':
			case 'smallint':
			case 'bigint':
			case 'smallserial':
			case 'serial':
			case 'bigserial':
				$col_info['type'] = 'int'; break;
			case 'float':
			case 'double':
			case 'real':
			case 'decimal':
			case 'numeric':
				$col_info['type'] = 'float'; break;
			default:
				$col_info['type'] = 'string'; break;
			}
		}
		unset($col_info);

		return $info;
	}

	/*!
	 *  \name Physical Devices
	 *  @{
	 */

	public function GetNumberOfPhysicalDevices($gid, $tid, $msid)
	{
		// format device filter statements, if necessary:
		$filter = array();
		if (!empty($gid)) {
			array_push($filter, 'hwgroup_id='.$gid);
		}
		if (!empty($tid)) {
			array_push($filter, 'device_type_id='.$tid);
		}
		if (!empty($msid)) {
			array_push($filter, 'model_spec_id='.$msid);
		}

		// query to filter physical devices:
		$sql = 'SELECT id FROM physical';
		if (count($filter) > 0) {
			$sql .= ' WHERE ' . implode(' AND ', $filter);
		}

		$ret = $this->Query($sql);
		if ($ret) {
			return count($ret);
		} else {
			return false;
		}
	}

	public function GetPhysicalDeviceId($gid, $tid, $sn)
	{
		$filter = array();

		array_push($filter, 'hwgroup_id='.$gid);
		array_push($filter, 'device_type_id='.$tid);
		array_push($filter,
			'serial_number=\''.$this->EscapeString($sn).'\'');

		// prepare query:
		$sql  = 'SELECT id FROM physical WHERE ';
		$sql .= implode(' AND ', $filter);

		$result = $this->Query($sql);
		if ($result !== false) {	// query OK, but maybe empty
			$nmatches = count($result);
			if ($nmatches > 1) {		// multiple matches
				$ret_val = -$nmatches;
			} else if ($nmatches == 1) {	// one match
				$ret_val = @pg_fetch_row($result)['id'];
			} else {			// no match
				$ret_val = 0;
			}
			return ($ret_val);
		} else {
			$cdata = 'physical: ' . print_r($sql, true);
			fatal($this->FmtLastError('Select', $cdata));
		}
	}

	/*!
	 *  end Physical Devices
	 *  @}
	 */

	protected function Query($sql)
	{
		$rows = array();

		if (!$this->mLink) {
			return false;
		}

		$result = pg_query($this->mLink, $sql);

		switch (@pg_result_status($result)) {
		case PGSQL_EMPTY_QUERY:
		case PGSQL_COMMAND_OK:
			return true;
		case PGSQL_TUPLES_OK:
			if (@pg_num_rows($result) > 0) {
				// retrieve results:
				while ($row = @pg_fetch_assoc($result)) {
					array_push($rows, $row);
				}
				$this->FixDatatypes($result, $rows);
			}
			return $rows;
		case PGSQL_BAD_RESPONSE:
		case PGSQL_FATAL_ERROR:
		default:
			fatal($this->FmtLastError('Query', $sql));
		}
	}

	/*!
	 *  Adjust values to PHP data types. Data returned by a query may not
	 *  always be compatible with PHP's interpretation of the data (e.g.,
	 *  Postgres' boolean data type). This function translates that data,
	 *  where necessary, such that it's appropriate for the corresponding
	 *  PHP data type.
	 *
	 *  @param[in] $result_rsc  Resource of database results.
	 *  @param[in] $result_rows Reference to an array holding the data that
	 *                          should be inspected and, where necessary,
	 *                          adjusted. This is a two-dimensional array
	 *                          where each element (index) holds another
	 *                          array,which is an associative array. The
	 *                          latter contains one row of the query
	 *                          results. A valid association must exist
	 *                          between the data and the database resource.
	 *
	 *  @todo
	 *  Alternate solution; preferrably one that eliminates this function.
	 */
	private function FixDatatypes($result_rsc, &$result_rows)
	{
		foreach ($result_rows as &$row) {
			foreach ($row as $k => &$val) {
				$num = @pg_field_num($result_rsc, $k);
				if ($num >= 0) {
					$type = @pg_field_type($result_rsc,
									$num);
					switch ($type) {
					/*
					 * Postgres uses 't' for TRUE and 'f'
					 * for FALSE. This is translated to a
					 * PHP boolean value.
					 */
					case 'bool':
					case 'boolean':
						$val = ($val=='t' ? true:false);
						break;
					// no other translations yet...
					default:
						break;
					}
				} // unknown fields are silently ignored
			}
			unset($val);
		}
		unset($row);
	}

	/*!
	 *  Retrieve untranslated information schema of table. This function
	 *  is the same as GetInfoSchema() except the table information is
	 *  not translated.
	 *
	 *  @sa
	 *  GetInfoSchema()
	 */
	private function GetInfoSchemaRaw($table)
	{
		$info = array();

		$sql  = 'SELECT column_name,data_type,character_maximum_length,column_default ';
		$sql .= 'FROM information_schema.columns ';
		$sql .= 'WHERE table_name=\'' . $table . '\'';
		$ret = $this->Query($sql);
		if ($ret) {
			foreach ($ret as $row) {
				$name = $row['column_name'];

				// prepare results in the proper format:
				$info[$name] = array(
					'type' => $row['data_type'],
					'max_length' => $row['character_maximum_length'],
					'default' => $row['column_default']
				      );
			}
		}
		return $info;
	}

	/*!
	 *  Return uptime of database.
	 *
	 *  @returns
	 *  String expressing uptime in years, months, days, hours, minutes,
	 *  and seconds. If years, months, or days equals zero then it's
	 *  omitted from the output string. So, the output is like this:
	 *  <pre></code>
	 *  [\<years\>, ][\<months\>, ][\<days\>, ]\<hours\>:\<min\>:\<sec\>
	 *  </code></pre>
	 *
	 *  An empty string is returned if the uptime could not be determined.
	 */
	private function GetUptime()
	{
		$uptime = '';

		$sub = 'SELECT age(now(),pg_postmaster_start_time()) AS uptime';
		$sql  = 'SELECT date_part(\'year\',  subq.uptime) AS y, ';
		$sql .= '       date_part(\'month\', subq.uptime) AS m, ';
		$sql .= '       date_part(\'day\',   subq.uptime) AS d, ';
		$sql .= '       date_part(\'hour\',  subq.uptime) AS h, ';
		$sql .= '       date_part(\'min\',   subq.uptime) AS i, ';
		$sql .= '       trunc(date_part(\'sec\', subq.uptime)) AS s ';
		$sql .= 'FROM ('. $sub .') AS subq';
		$ret = $this->Query($sql);
		if ($ret) {
			if ($ret[0]['y'] != 0)
				$uptime .= sprintf('%d years, ', $ret[0]['y']);
			if ($ret[0]['m'] != 0)
				$uptime .= sprintf('%d months, ', $ret[0]['m']);
			if ($ret[0]['d'] != 0)
				$uptime .= sprintf('%d days, ', $ret[0]['d']);
			$uptime .= sprintf('%02d:%02d:%02d',
						$ret[0]['h'],
						$ret[0]['i'],
						$ret[0]['s']);
		}
		return $uptime;
	}

	const DB_IMPL = 'PostgreSQL';
}
?>
