<?php
/*!
 *  @file
 *
 *  @brief	implementation of generic SQL database access
 *
 *  @date	July 4, 2015
 */

/*!
 *  Implementation of SQL database access for HWS. This class defines
 *  functions, properties, and queries which are not specific to a certain
 *  database implementation.
 */
abstract class SqlDatabase
{
	/*!
	 *  Create database object. After the object has been created, no
	 *  connection to the database host is set up, yet.
	 *
	 *  @param[in] $host Host serving database. The name of the host
	 *                   may be followed by a colon and a port number.
	 *  @param[in] $user Name of user for login.
	 *  @param[in] $pass Unencrypted password of user logging in.
	 */
	function __construct($host, $user, $pass)
	{
		$this->mHost = $host;
		$this->mUser = $user;
		$this->mPass = $pass;
		$this->mDatabase = '';
		$this->mLink = false;
	}

	/*!
	 *  Destructor.
	 */
	function __destruct()
	{
	}

	/*!
	 *  Return database information. A list of administrative information
	 *  about the database being used is returned by this function. This
	 *  includes the following information:
	 *
	 *  - implementation
	 *  - host
	 *  - state
	 *  - version
	 *  - uptime
	 *
	 *  @returns
	 *  The database information is returned in an associative array. The
	 *  keys of the array are as listed above.
	 */
	abstract public function GetDbInfo();

	/*!
	 *  Create new database.
	 *
	 *  @param[in] $db      Name of database.
	 *  @param[in] $options Database-specific options (optional).
	 *
	 *  @returns
	 *  @c true in case the given database was created; @c false is
	 *  returned in case an error occurred. Note that it's also considered
	 *  an error if the database already exists.
	 */
	public function CreateDatabase($db, $options = '')
	{
		if (trim($db) == false) {
			return false;
		}

		// format query:
		$sql = 'CREATE DATABASE ' . $db . ' ' . $options;
		return $this->Query($sql);
	}

	/*!
	 *  Create new database table. Before a table can be created, a
	 *  connection must have been set-up with a (existing) database.
	 *
	 *  @param[in] $table    Name of table.
	 *  @param[in] $col_defs List of column definitions for the table.
	 *                       The list consists of an index-based array
	 *                       where each element holds an column
	 *                       definition. The elements in the array, and
	 *                       consequently the column definitions, are
	 *                       processed in ascending order.
	 *  @param[in] $options  Table-specific options (optional).
	 *
	 *  @returns
	 *  Boolean value indicating whether the table was created
	 *  successfully. If the table did already exist then this is
	 *  considered an error.
	 */
	public function CreateTable($table, $col_defs, $options = '')
	{
		if (trim($table) == false) {
			return false;
		}

		// format query:
		$sql = 'CREATE TABLE ' . $table . ' ';
		if (count($col_defs) > 0) {
			$sql .= '(' . implode(',', $col_defs) . ')';
		}
		$sql .= $options;

		return $this->Query($sql);
	}

	/*!
	 *  Create trigger for database table. The function only supports
	 *  simple trigger definitions.
	 *
	 *  @param[in] $trigger Name of trigger.
	 *  @param[in] $table   Table associated with trigger.
	 *  @param[in] $event   Event causing the trigger to fire. The
	 *                      following events are supported:
	 *                      - @c INSERT
	 *                      - @c UPDATE
	 *                      - @c DELETE
	 *  @param[in] $when    Specifies when the trigger should fire
	 *                      in relation to the event. This parameter
	 *                      can take one of the following values:
	 *                      - @c BEFORE
	 *                      - @c AFTER
	 *  @param[in] $stmts   Array holding one or more SQL statements
	 *                      which are executed whenever the trigger fires.
	 *                      Each element of the array holds an SQL
	 *                      statement. The semi-colon delimiter is
	 *                      automatically added for each statement.
	 *
	 *  @returns
	 *  Boolean value indicating whether the trigger was created
	 *  successfully or not.
	 */
	function CreateTrigger($trigger, $table, $event, $when, $stmts)
	{
		if (trim($trigger) == false ||
		    trim($table)   == false)
		{
			return false;
		}

		$ev = strtoupper(trim($event));
		if (strcmp($ev, 'INSERT') != 0 &&
		    strcmp($ev, 'UPDATE') != 0 &&
		    strcmp($ev, 'DELETE') != 0)
		{
			return false;
		}

		$wh = strtoupper(trim($when));
		if (strcmp($wh, 'BEFORE') != 0 &&
		    strcmp($wh, 'AFTER')  != 0)
		{
			return false;
		}

		// format query:
		$sql  = 'CREATE TRIGGER ' . $trigger . ' ' . $wh . ' ' . $ev;
		$sql .= ' ON ' . $table;
		$sql .= ' BEGIN ' . implode(';', $stmts) . '; END';
		return $this->Query($sql);
	}

	/*!
	 *  Setup connection to database host and set database to use. If no
	 *  database is specified, then only a connection to the DBMS is
	 *  attempted.
	 *
	 *  @param[in] $db Name of database.
	 *
	 *  @returns
	 *  @c true if the connection has been established and the database
	 *  is ready for use; @c false in case an error occurred.
	 */
	abstract public function Connect($db);

	/*!
	 *  Close database connection.
	 */
	abstract public function Close();

	/*!
	 *  Add given set of values to database. The set of values is passed
	 *  as an associative array holding the column names and values for
	 *  1 row in the given table.
	 *
	 *  @param[in] $table  Name of table.
	 *  @param[in] $fields Associative array where the @e key indicates
	 *                     the column name in the table and the @e value
	 *                     holds the value for that column.
	 *
	 *  @returns
	 *  ID of inserted row if successful, otherwise @c false.
	 */
	abstract public function Insert($table, $fields);

	/*!
	 *  Update values of given row in database. The row to be updated by
	 *  this function is given by an ID number which must be part of the
	 *  definition of a table. ID numbers are assumed to be stored in a
	 *  column named @c id.
	 *
	 *  The new values of the row are passed by means of an associative
	 *  array which holds the column names and their value.
	 *
	 *  @param[in] $table  Name of table.
	 *  @param[in] $id     ID number of row to update.
	 *  @param[in] $fields Associative array where the @e key indicates
	 *                     the column name in the table and the @e value
	 *                     holds the value for that column.
	 *
	 *  @returns
	 *  @c true is returned if the values have been updated; @c false is
	 *  returned if the update failed.
	 */
	abstract public function Update($table, $id, $fields);

	/*!
	 *  Execute SQL query on database.
	 *
	 *  @param[in] $sql Query to be executed. The query string is @e not
	 *                  scanned for special characters, so the caller must
	 *                  ensure there are no harmful characters in the
	 *                  query. Also, the query should not contain the
	 *                  semicolon to terminate the statement.
	 *
	 *  @returns
	 *  Dependent on the query. The function may return an array with
	 *  results data, or @c true if the query succeeded. This depends
	 *  on the query being executed. @c false is returned in case an error
	 *  occurred. However, a @e fatal error occurs if the execution of the
	 *  query itself failed.
	 *
	 *  @remarks
	 *  This function is protected to prevent exposing details of the
	 *  underlying database implementation which could become apparent
	 *  by certain SQL queries.
	 */
	abstract protected function Query($sql);

	/*!
	 *  Retrieve given columns of all rows in database table. This
	 *  function selects one or more columns for each row in the given
	 *  table and returns the result. The results can optionally be
	 *  sorted on a specific column prior to being returned.
	 *
	 *  @param[in] $table   Name of table.
	 *  @param[in] $columns List of columns names. The names are stored
	 *                      in an array where each element indicates a
	 *                      column name. In case no columns are given,
	 *                      the function selects all columns.
	 *  @param[in] $order   Column name to use for sorting the results.
	 *                      This parameter may be left blank.
	 *
	 *  @returns
	 *  Two-dimensional array where each element (index) represents one
	 *  row of the results. An element of the array holds another
	 *  (associative) array which holds the retrieved columns and values.
	 *  In case an error occurred, the function will return @c false.
	 */
	public function Fetch($table, $columns, $order = '')
	{
		if (!$this->mLink) {
			return false;
		}

		if (count($columns) == 0) {
			$col_names = '*';
		} else {
			$col_names = implode(',', $columns);
		}

		// format query:
		$sql = 'SELECT ' . $col_names . ' FROM ' . $table;
		if (trim($order)) {
			$sql .= ' ORDER BY ' . $order;
		}
		return $this->Query($sql);
	}

	/*!
	 *  Retrieve columns of given row in database table. The row in the
	 *  table is specified by an ID number which must be part of the table
	 *  definition. ID numbers are assumed to be stored in a column named
	 *  @c id.
	 *
	 *  @param[in] $table   Name of table.
	 *  @param[in] $columns List of columns names. The names are stored
	 *                      in an array where each element indicates a
	 *                      column name. In case no columns are given,
	 *                      the function will select all columns.
	 *  @param[in] $id      ID number of row to fetch.
	 *
	 *  @returns
	 *  Associative array holding the column names and their value (as
	 *  key/value pairs). @c false is returned in case an error occurred.
	 */
	public function FetchById($table, $columns, $id)
	{
		if (!$this->mLink || trim($id) == false) {
			return false;
		}

		if (count($columns) == 0) {
			$col_names = '*';
		} else {
			$col_names = implode(',', $columns);
		}

		// format query:
		$sql = 'SELECT '.$col_names.' FROM '.$table.' WHERE id='.$id;
		$result = $this->Query($sql);
		if ($result) {
			// there should be only 1 row:
			return $result[0];
		} else {
			return false;
		} 
	}

	/*!
	 *  Retrieve rows matching column and value. For a given table, this
	 *  function takes a list of column/value pairs and returns all rows
	 *  where the column(s) match the associated value.
	 *
	 *  @param[in] $table  Name of table.
	 *  @param[in] $values List of column/value pairs, stored in an
	 *                     associative array. The @e key of the array
	 *                     corresponds to a column in the table, while the
	 *                     @e value corresponds to the value of the said
	 *                     column. The value is always quoted before it is
	 *                     inserted into the query. In case no column/value
	 *                     pairs are given, then a row will always match.
	 *                     Effectively, this will retrieve the entire table.
	 *
	 *  @param[in] $order  Column name to use for sorting the results.
	 *                     This parameter may be left blank.
	 *
	 *  @returns
	 *  Two-dimensional array where each element (index) represents one
	 *  row of the results. An element of the array holds another
	 *  (associative) array which holds the retrieved columns and values.
	 *  In case an error occurred, the function will return @c false.
	 */
	public function FetchByValue($table, $values, $order = '')
	{
		if (!$this->mLink) {
			return false;
		}

		if (!empty($values)) {
			$expr = array();
			foreach ($values as $k => $v) {
				$tmp = $this->FmtSqlParameter($table, $k, $v);
				array_push($expr, "$k=$tmp");
			}
			$where = 'WHERE ' . implode(' AND ', $expr);
		}

		// format query:
		$sql = 'SELECT * FROM ' . $table . ' ' . $where;
		if (trim($order)) {
			$sql .= ' ORDER BY ' . $order;
		}
		return $this->Query($sql);
	}

	/*!
	 *  Retrieve rows matching column and value. For a given table, this
	 *  function takes the name of a column and a list of values. Rows
	 *  where the given column matches one of the values are returned.
	 *
	 *  @param[in] $table  Name of table.
	 *  @param[in] $column Column name to use.
	 *  @param[in] $values List of values, stored in an array. The value
	 *                     is always quoted before it is inserted into the
	 *                     query. Empty elements in $values are ignored.
	 *                     However, in case $values is completely empty,
	 *                     then a row of the table always matches, so
	 *                     the entire table will be returned.
	 *
	 *  @returns
	 *  Two-dimensional array where each element (index) represents one
	 *  row of the results. An element of the array holds another
	 *  (associative) array which holds the retrieved columns and values.
	 *  In case an error occurred, the function will return @c false.
	 */
	public function FetchByMultiValue($table, $column, $values)
	{
		if (!$this->mLink) {
			return false;
		}

		if (!empty($values)) {
			$expr = array();
			foreach ($values as $v) {
				if (empty($v)) continue;

				$tmp = $this->FmtSqlParameter($table, $column, $v);
				array_push($expr, "$column=$tmp");
			}
			$where = 'WHERE ' . implode(' OR ', $expr);
		}

		// format query:
		$sql = 'SELECT * FROM ' . $table . ' ' . $where;
		return $this->Query($sql);
	}

	/*!
	 *  Return escaped string for query. The function escapes harmful
	 *  characters in the given string such that the converted string
	 *  can be used safely in an SQL query.
	 *
	 *  @param[in] $str Unescaped string.
	 *
	 *  @returns
	 *  String with harmful characters escaped.
	 */
	abstract public function EscapeString($str);

	/*!
	 *  Translate value to SQL parameter. The given value is examined and
	 *  formatted such that it can be used as an SQL parameter (i.e.,
	 *  inserted in a query). How the parameter is translated depends on
	 *  the data type of the value and on the database in question. For
	 *  instance, a value may only need quoting or it may be represented
	 *  by a special keyword.
	 *
	 *  Note that the value undergoes escaping before it's returned.
	 *
	 *  @param[in] $table  Name of table the value is stored in.
	 *  @param[in] $column Value is stored in this column.
	 *  @param[in] $value  Value to translate.
	 *
	 *  @returns
	 *  SQL parameter. If the value does not represent a scalar data type
	 *  (e.g., an array), the function returns @c false.
	 *
	 *  @note
	 *  Any value not recognized is returned unmodified.
	 */
	abstract public function FmtSqlParameter($table, $column, $value);

	/*!
	 *  Format message for last error occurred. The function formats a
	 *  string consisting of the following items:
	 *
	 *  - Error of last database operation
	 *  - Name of the database
	 *  - Database backend identification
	 *  - User-defined ID for this error ("context")
	 *  - User-defined text associated with context
	 *
	 *  @param[in] $context Identification for this error (preferably
	 *                      a single word).
	 *  @param[in] $data    Additional text relevant for this error.
	 *
	 *  @returns
	 *  Error string.
	 */
	abstract public function FmtLastError($context, $data);

	/*!
	 *  Retrieve information schema of table. The returned information
	 *  describes (part of) the structure of the given table. For each
	 *  column in the table the following information is returned:
	 *
	 *  - Column name
	 *  - Data type
	 *  - Maximum length (only character types)
	 *  - Default value
	 *
	 *  If necessary, the returned table information is translated to
	 *  PHP-equivalents. For instance, the data type is adapted such
	 *  that it reflects a PHP data type.
	 *
	 *  @param[in] $table Name of table.
	 *
	 *  @returns
	 *  Array where each row holds a second array which contains the
	 *  description of a column. Both arrays are associative arrays
	 *  where the array keys are named after the column information
	 *  mentioned above. The layout of the arrays is as shown below.
	 *  <pre><code>
	 *  [column] => (
	 *                [type]       => \<value\>
	 *                [max_length] => \<value\>
	 *                [default]    => \<value\>
	 *              )
	 *  </code></pre>
	 */
	abstract public function GetInfoSchema($table);

	/*!
	 *  \name Physical Devices
	 *  @{
	 */

	/*!
	 *  Determine amount of physical devices. An ID number can be given
	 *  to filter on a device's hardware group, device type, and/or
	 *  its model-specific ID. The ID numbers are optional.
	 *
	 *  @param[in] $gid  Hardware group ID.
	 *  @param[in] $tid  Device type ID.
	 *  @param[in] $msid Model-specific ID within device type.
	 *
	 *  @returns
	 *  Non-negative number indicating the amount of matching devices or
	 *  @c false on error.
	 */
	abstract public function GetNumberOfPhysicalDevices($gid, $tid, $msid);

	/*!
	 *  Determine ID of physical device.
	 *
	 *  @param[in] $gid Hardware group ID.
	 *  @param[in] $tid Device type ID.
	 *  @param[in] $sn  Serial number of device.
	 *
	 *  @returns
	 *  Integer value indicating result of operation:
	 *  - <0: negative value indicates multiple matches; the value
	 *        indicates how many
	 *  - =0: no ID found
	 *  - >0: positive value indicates one match; value indicates ID
	 */
	abstract public function GetPhysicalDeviceId($gid, $tid, $sn);

	/*!
	 *  Retrieve all devices of given group, type, and model. The
	 *  resulting table lists the ID of the device(s) along with their
	 *  serial number, status, and location.
	 *  <pre><code>
	 *  +----+----+---------+--------+----------+
	 *  | id | sn | barcode | status | location |
	 *  +----+----+---------+--------+----------+
	 *  </code></pre>
	 *
	 *  @param[in] $gid   Hardware group ID.
	 *  @param[in] $tid   Device type ID.
	 *  @param[in] $msid  Model-specific ID within device type.
	 *  @param[in] $order Column name for sorting the results.
	 *
	 *  @returns
	 *  Two-dimensional array where each element (index) represents one
	 *  row of the results. An element of the array holds another
	 *  (associative) array which holds the columns depicted above for
	 *  a device. In case an error occurred, the function will return
	 *  @c false.
	 */
	public function ListDevices($gid, $tid, $msid, $order)
	{
		if (trim($gid)  == false ||
		    trim($tid)  == false ||
		    trim($msid) == false)
		{
			return false;
		}

		// format query:
		$sql  = 'SELECT p.id,p.serial_number,p.barcode,s.status,l.location ';
		$sql .= 'FROM physical p ';
		$sql .= 'LEFT JOIN status    s ON s.id=p.status_id ';
		$sql .= 'LEFT JOIN locations l ON l.id=p.location_id ';
		$sql .= 'WHERE p.hwgroup_id=' . $gid;
		$sql .= '  AND p.device_type_id=' . $tid;
		$sql .= '  AND p.model_spec_id=' . $msid;
		$sql .= ' ORDER BY ' . $order;
		return $this->Query($sql);
	}

	/*!
	 *  Retrieve state of device. The resulting table lists IDs and the
	 *  state of the device as depicted in the table below.
	 *  <pre><code>
	 *  +------------+---------------+----------------+---------------+~
	 *  | hwgroup_id | model_spec_id | device_type_id | serial_number |
	 *  +------------+---------------+----------------+---------------+~
	 *                                     ~---------+--------+----------+
	 *                                       barcode | status | location |
	 *                                     ~---------+--------+----------+
	 *  </code></pre>
	 *
	 *  @param[in] $did Physical device ID.
	 *
	 *  @returns
	 *  Associative array holding column names depicted above (as @e keys)
	 *  and their values. In case of an error, @c false is returned.
	 */
	public function GetDeviceState($did)
	{
		if (trim($did) == false) {
			return false;
		}

		// format query:
		$sql  = 'SELECT p.hwgroup_id,p.model_spec_id,';
		$sql .= '       p.device_type_id,p.serial_number,';
		$sql .= '       p.barcode,s.status,l.location ';
		$sql .= 'FROM physical p ';
		$sql .= 'LEFT JOIN status    s ON s.id=p.status_id ';
		$sql .= 'LEFT JOIN locations l ON l.id=p.location_id ';
		$sql .= 'WHERE p.id=' . $did;
		$result = $this->Query($sql);
		if ($result) {
			return $result[0];
		} else {
			return false;
		}
	}

	/*!
	 *  Retrieve changes of device. A table holding (basic) changes made
	 *  to one or more devices is returned. The following columns will be
	 *  returned for each matching device:
	 *  <pre><code>
	 *  +-------------+------+------+---------------+--------+------+
	 *  | physical_id | date | time | serial_number | action | user |
	 *  +-------------+------+------+---------------+--------+------+
	 *  </code></pre>
	 *
	 *  @param[in] $did   Physical device ID. This parameter is optional
	 *                    and if omitted, the function returns the changes
	 *                    of all physical devices.
	 *  @param[in] $limit Limitting number indicating how many devices to
	 *                    return.
	 *  @param[in] $asc   Boolean value controlling whether to sort in
	 *                    ascending order (@c true) or descending order
	 *                    (@c false). The results are always sorted on
	 *                    date and time. The default for this parameter
	 *                    is @c true.
	 *
	 *  @returns
	 *  Two-dimensional array where each element (index) represents one
	 *  row of the results. An element of the array holds another
	 *  (associative) array which holds the columns depicted above for
	 *  a device. In case an error occurred, the function will return
	 *  @c false.
	 */
	public function GetDeviceChanges($did, $limit, $asc = true)
	{
		$order = ($asc === true ? 'ASC' : 'DESC');
	
		// format query:
		$sql  = 'SELECT h.physical_id,h.date,h.time,p.serial_number, ';
		$sql .= '       h.action,u.username ';
		$sql .= 'FROM physical_history h ';
		$sql .= 'LEFT JOIN physical p ON p.id=h.physical_id ';
		$sql .= 'LEFT JOIN users u    ON u.id=h.user_id ';
		if (!empty($did)) {
			$sql .= 'WHERE h.physical_id=' . $did . ' ';
		}
		$sql .= 'ORDER BY date ' . $order . ', time ' . $order . ' ';
		if (!empty($limit)) {
			$sql .= 'LIMIT ' . $limit;
		}
		return $this->Query($sql);
	}

	/*!
	 *  Return note IDs associated with device.
	 *
	 *  @param[in] $did Physical device ID.
	 *
	 *  @returns
	 *  List of ID numbers of the notes associated with the given device.
	 *  The list is in the form of an array where each ID is stored as an
	 *  element of the array.
	 */
	public function GetDeviceNotesById($did)
	{
		$ids = array();

		if (trim($did) == false) {
			return $ids;
		}

		$sql = 'SELECT id FROM notes WHERE physical_id=' . $did;
		$ret = $this->Query($sql);
		// results are in a 2-dimensional array

		// save matching IDs in 1-dimensional array:
		foreach ($ret as $row) {
			array_push($ids, $row['id']);
		}

		return $ids;
	}

	/*!
	 *  Return link IDs associated with device.
	 *
	 *  @param[in] $did Physical device ID.
	 *
	 *  @returns
	 *  List of ID numbers of the links associated with the given device.
	 *  The list is in the form of an array where each element of the
	 *  array holds a link ID.
	 */
	public function GetDeviceLinksById($did)
	{
		$ids = array();

		if (trim($did) == false) {
			return $ids;
		}

		$sql = 'SELECT id FROM links WHERE physical_id=' . $did;
		$ret = $this->Query($sql);
		// results are in a 2-dimensional array

		// save matching IDs in 1-dimensional array:
		foreach ($ret as $row) {
			array_push($ids, $row['id']);
		}

		return $ids;
	}

	/*!
	 *  Remove given physical device. Notes or links associated with the
	 *  device are not removed.
	 *
	 *  @param[in] $did Physical device ID.
	 *
	 *  @returns
	 *  @c true is returned if device has been removed, otherwise @c false.
	 */
	public function DeleteDevice($did)
	{
		if (trim($did) == false) {
			return false;
		}

		$sql = 'DELETE FROM physical WHERE id=' . $did;

		return $this->Query($sql);
	}

	/*!
	 *  end Physical Devices
	 *  @}
	 *
	 *  \name Device Types
	 *  @{
	 */

	/*!
	 *  Return list of device types with given hardware group. A table is
	 *  returned where each row indicates the name of a device type and
	 *  its corresponding ID number (see below). The table is sorted
	 *  alphabetically (on device type).
	 *  <pre><code>
	 *  +----+-------------+
	 *  | id | device_type |
	 *  +----+-------------+
	 *  </code></pre>
	 *
	 *  @param[in] $gid Hardware group ID.
	 *
	 *  @returns
	 *  Two-dimensional array where each element (index) represents one
	 *  row of the results. An element of the array holds another
	 *  (associative) array which holds the columns depicted above for a
	 *  device type. In case an error occurred, the function will return
	 *  @c false.
	 */
	public function ListDeviceTypes($gid)
	{
		if (trim($gid) == false) {
			return false;
		}

		$sql  = 'SELECT id,device_type FROM device_type ';
		$sql .= 'WHERE hwgroup_id='.$gid.' ORDER BY device_type';

		return $this->Query($sql);
	}

	/*!
	 *  Return shortcode for given device type.
	 *
	 *  @param[in] $tid Device type ID.
	 *
	 *  @returns
	 *  Shortcode associated with device type or @c false in case of
	 *  an error. Note that empty shortcode or a shortcode consisting
	 *  of whitespace only, is also considered an error.
	 */
	public function GetDeviceTypeCode($tid)
	{
		$code = '';

		$results = $this->FetchById('device_type', array('code'), $tid);
		if ($results) {
			$code = trim($results['code']);
		}
		return (!empty($code) ? $code : false);
	}

	/*!
	 *  Return description for given device type.
	 *
	 *  @param[in] $tid Device type ID.
	 *
	 *  @returns
	 *  String holding description of device type or @c false in case of
	 *  an error.
	 */
	public function GetDeviceTypeDesc($tid)
	{
		$desc = '';

		$ret = $this->FetchById('device_type',
					array('device_type'),
					$tid);
		if ($ret) {
			$desc = trim($ret['device_type']);
		}
		return (!empty($desc) ? $desc : false);
	}

	/*!
	 *  Return list of models of given device type. This function takes
	 *  the table of the given device type and substitutes the appropriate
	 *  string value for each feature. In principle the features in the
	 *  table are stored by ID number, so the function will replace these
	 *  numbers by their corresponding string value. The resulting table
	 *  lists all available models for the given device type in textual
	 *  form.
	 *  <pre><code>
	 *  +----+-------+-------------+-----+
	 *  | id | model | manufacture | ... |
	 *  +----+-------+-------------+-----+
	 *  </code></pre>
	 *
	 *  The resulting table is highly dependent on the device type since
	 *  different device types have different features. However, the three
	 *  columns depicted above are common to all device types.
	 *
	 *  @param[in] $mstable      Name of model-specific table.
	 *  @param[in] $feature_defs Feature definitions which specify the
	 *                           features for the device type. See the DTD
	 *                           files for more information.
	 *  @param[in] $order        Sort results on column specified by this
	 *                           parameter. If omitted, the results are
	 *                           sorted on model name.
	 *
	 *  @returns
	 *  Two-dimensional array where each element (index) represents one
	 *  row of the results. An element of the array holds another
	 *  (associative) array which holds the columns depicted above for
	 *  a device type. In case an error occurred, the function will return
	 *  @c false.
	 */
	public function ListModels($mstable, $feature_defs, $order = 'model')
	{
		if (empty($feature_defs)) {
			return false;
		}

		/*
		 * Using a separate query for the device type features. This
		 * query joins multiple tables together, but some tables may
		 * have columns with the same name (i.e., "id"). When using
		 * such a column with the ORDER BY statement, some databases
		 * (e.g., SQLite3) see this as an ambiguity and the whole
		 * query fails. Therefore, we use a "sub-query" on which the
		 * ORDER BY is performed.
		 */
		$sub = $this->FormatDeviceTypeFeatures($mstable, $feature_defs);
		$sql = 'SELECT * FROM ('.$sub.') AS dummy ORDER BY '.$order;

		return $this->Query($sql);
	
	}

	/*!
	 *  Return features of given model. This function resembles the
	 *  function ListModels() except only the given model is returned.
	 *
	 *  @param[in] $mstable      Name of model-specific table.
	 *  @param[in] $feature_defs Feature definitions which specify the
	 *                           features for the device type. See the DTD
	 *                           files for more information.
	 *  @param[in] $msid         Model-specific ID within device type.
	 *
	 *  @returns
	 *  If model could be found, an associative array holding its
	 *  features is returned. If not found, @c null is returned.
	 */
	public function GetModelFeatures($mstable, $feature_defs, $msid)
	{
		if (trim($msid) == false) {
			return null;
		}

		$results = $this->ListModels($mstable, $feature_defs, 'id');
		foreach ($results as $row) {
			if ($row['id'] == $msid) {
				return $row;
			}
		}
		return null;
	}

	/*!
	 *  end Device Types
	 *  @}
	 *
	 *  \name Device Notes and Links
	 *  @{
	 */

	/*!
	 *  Remove given note associated with a physical device.
	 *
	 *  @param[in] $nid ID number of note.
	 *
	 *  @returns
	 *  @c true is returned if note has been removed, otherwise @c false.
	 */
	public function DeleteNote($nid)
	{
		if (trim($nid) == false) {
			return false;
		}

		$sql = 'DELETE FROM notes WHERE id=' . $nid;

		return $this->Query($sql);
	}
	
	/*!
	 *  Remove given URL/link associated with a physical device.
	 *
	 *  @param[in] $lid ID number of link.
	 *
	 *  @returns
	 *  @c true is returned if link has been removed, otherwise @c false.
	 */
	public function DeleteLink($lid)
	{
		if (trim($lid) == false) {
			return false;
		}

		$sql = 'DELETE FROM links WHERE id=' . $lid;

		return $this->Query($sql);
	}

	/*!
	 *  end Device Notes and Links
	 *  @}
	 *
	 *  \name User Accounts
	 *  @{
	 */

	/*!
	 *  Add new user account. A user must (obviously) have a name, but a
	 *  password is not required. Also, the permissions may be omitted
	 *  though this doesn't seem to make sense.
	 *
	 *  @param[in] $user     Name of (new) user.
	 *  @param[in] $pass     Password of user. The password can be stored
	 *                       encrypted or unencrypted; how the password is
	 *                       interpreted is determined by the caller.
	 *                       However, encrypted passwords should be encode
	 *                       by the caller such that the @c $pass does not
	 *                       contain special characters.
	 *  @param[in] $perm     Access permissions of user.
	 *  @param[in] $fullname Full name of this user (optional).
	 *  @param[in] $locked   Indicates account is locked or not. By default
	 *                       the account is not locked.
	 *
	 *  @returns
	 *  ID number of new user account if successful; @c false in case
	 *  the account could not be created.
	 */
	public function AddUser($user, $pass, $perm, $fullname, $locked = false)
	{
		if (trim($user) == false) {
			return false;
		}

		$fields = array('username'    => $user,
				'permissions' => $perm,
				'password'    => $pass,
				'fullname'    => $fullname,
				'locked'      => $locked);

		return $this->Insert('users', $fields);
	}

	/*!
	 *  Remove given user account.
	 *
	 *  @param[in] $uid ID number of user account.
	 *
	 *  @returns
	 *  @c true is returned if user account has been removed, otherwise
	 *  @c false.
	 */
	public function DeleteUser($uid)
	{
		if (trim($uid) == false) {
			return false;
		}

		$sql = 'DELETE FROM users WHERE id=' . $uid;

		return $this->Query($sql);
	}

	/*!
	 *  Get ID number associated with user.
	 *
	 *  @param[in] $user Name of user.
	 *
	 *  @returns
	 *  Positive integer indicating ID number of user or 0 on failure.
	 *
	 *  @note
	 *  In case the user matches multiple ID numbers, then only the
	 *  first matching ID number is returned.
	 */
	public function GetUserId($user)
	{
		if (trim($user) == false) {
			return 0;
		}

		$sql = 'SELECT id FROM users WHERE username=\''.$user.'\'';
		$ret = $this->Query($sql);
		if ($ret) {
			return $ret[0]['id'];
		} else {
			return 0;
		}
	}

	/*!
	 *  Retrieve user account information.
	 *
	 *  @param[in] $uid ID number of user account.
	 *
	 *  @returns
	 *  Associative array holding account information if successful;
	 *  @c false on error or unknown user.
	 */
	public function GetUserInfo($uid)
	{
		return $this->FetchById('users', array(), $uid);
	}

	/*!
	 *  Change a user account.
	 *
	 *  @param[in] $uid        ID number of user account.
	 *  @param[in] $properties Associative array holding key/value pairs
	 *                         of the account properties to be modified.
	 *
	 *  @returns
	 *  Boolean value indicating whether all modifications have be stored.
	 */
	public function ModUserInfo($uid, $properties)
	{
		if (trim($uid) == false) {
			return false;
		}

		return $this->Update('users', $uid, $properties);
	}

	/*!
	 *  end User Accounts
	 *  @}
	 *
	 *  \name Location Definitions
	 *  @{
	 */

	/*!
	 *  Add new location. A location must have an identifying name, but
	 *  does not require any additional details.
	 *
	 *  @param[in] $location Identifier of (new) location.
	 *  @param[in] $loc_info Associative array holding the details of
	 *                       the location. The following items (@e keys)
	 *                       can be specified in the array:
	 *
	 *                       - street
	 *                       - zipcode
	 *                       - city
	 *                       - room
	 *                       - rack
	 *                       - shelf
	 *
	 *                       Note that these keys are case-sensitive.
	 *
	 *  @returns
	 *  ID number of new location if successful; @c false in case
	 *  the location could not be created.
	 */
	public function AddLocation($location, $loc_info)
	{
		if (trim($location) == false) {
			return false;
		}

		$loc_info['location'] = $location;

		return $this->Insert('locations', $loc_info);
	}

	/*!
	 *  Remove given location.
	 *
	 *  @param[in] $lid ID number of location definition.
	 *
	 *  @returns
	 *  @c true is returned if location has been removed, otherwise
	 *  @c false.
	 */
	public function DeleteLocation($lid)
	{
		if (trim($lid) == false) {
			return false;
		}

		$sql = 'DELETE FROM locations WHERE id=' . $lid;

		return $this->Query($sql);
	}

	/*!
	 *  Get ID number associated with location.
	 *
	 *  @param[in] $location Name of location definition.
	 *
	 *  @returns
	 *  Positive integer indicating ID number of location or 0 on failure.
	 *
	 *  @note
	 *  In case the location matches multiple ID numbers, then only the
	 *  first matching ID number is returned.
	 */
	public function GetLocationId($location)
	{
		if (trim($location) == false) {
			return 0;
		}

		$sql = 'SELECT id FROM locations WHERE location=\''.$location.'\'';
		$ret = $this->Query($sql);
		if ($ret) {
			return $ret[0]['id'];
		} else {
			return 0;
		}
	}

	/*!
	 *  Retrieve location information.
	 *
	 *  @param[in] $lid ID number of location definition.
	 *
	 *  @returns
	 *  Associative array holding location information if successful;
	 *  @c false on error or unknown location.
	 */
	public function GetLocationInfo($lid)
	{
		return $this->FetchById('locations', array(), $lid);
	}

	/*!
	 *  Change a location definition.
	 *
	 *  @param[in] $lid        ID number of location definition.
	 *  @param[in] $properties Associative array holding key/value pairs
	 *                         of the location properties to be modified.
	 *
	 *  @returns
	 *  Boolean value indicating whether all modifications have be stored.
	 */
	public function ModLocationInfo($lid, $properties)
	{
		if (trim($lid) == false) {
			return false;
		}

		return $this->Update('locations', $lid, $properties);
	}

	/*!
	 *  end Location Definitions
	 *  @}
	 *
	 *  \name News Items
	 *  @{
	 */

	/*!
	 *  Retrieve one or more news items. The number of items to retrieve is
	 *  determined by a parameter and for each item that is retrieved, the
	 *  following columns are returned:
	 *  <pre><code>
	 *  +----+------+------+---------+---------+
	 *  | id | date | time | subject | message |
	 *  +----+------+------+---------+---------+
	 *  </code></pre>
	 *
	 *  The news items retrieved are always sorted by date and time.
	 *
	 *  @param[in] $limit Limitting number indicating how many news
	 *                    items to return.
	 *  @param[in] $asc   Boolean value controlling whether to sort in
	 *                    ascending order (@c true) or descending order
	 *                    (@c false). The default for this parameter
	 *                    is @c false.
	 *
	 *  @returns
	 *  Two-dimensional array where each element (index) represents one
	 *  row of the results. An element of the array holds another
	 *  (associative) array which holds the columns depicted above for
	 *  a device. In case an error occurred, the function will return
	 *  @c false.
	 */
	public function FetchNews($limit, $asc = false)
	{
		$order = ($asc === true ? 'ASC' : 'DESC');

		// format query:
		$sql  = 'SELECT id,date,time,subject,message FROM news ';
		$sql .= 'ORDER BY date ' . $order . ', time ' . $order . ' ';
		if (!empty($limit)) {
			$sql .= 'LIMIT ' . $limit;
		}
		return $this->Query($sql);
	}

	/*!
	 *  Remove given news item.
	 *
	 *  @param[in] $nid ID number of news item.
	 *
	 *  @returns
	 *  @c true is returned if item has been removed, otherwise @c false.
	 */
	public function DeleteNews($nid)
	{
		if (trim($nid) == false) {
			return false;
		}

		$sql = 'DELETE FROM news WHERE id=' . $nid;

		return $this->Query($sql);
	}

	/*!
	 *  end News Items
	 *  @}
	 */

	/*!
	 *  Format query for listing features of device type. The resulting
	 *  query has a form as shown below. <code>table</code>,
	 *  <code>mstable</code>, and <code>col_name</code> are placeholders
	 *  for resp. the device type table, a model-specific table, and a
	 *  column in a model-specific table.
	 *  <pre><code>
	 *  SELECT y.id,y.model_id,v0.\<col_name0\>,v1.\<col_name1\>,...
	 *  FROM \<table\> y
	 *  LEFT JOIN \<mstable0\> v0 ON v0.id=y.\<col_name0\>_id,
	 *  LEFT JOIN \<mstable1\> v1 ON v1.id=y.\<col_name1\>_id,
	 *  ...
	 *  </code></pre>
	 *
	 *  @param[in] $mstable      Name of model-specific table.
	 *  @param[in] $feature_defs Feature definitions which specify the
	 *                           features for the device type. See the DTD
	 *                           files for more information.
	 *
	 *  @returns
	 *  Query string.
	 */
	private function FormatDeviceTypeFeatures($mstable, $feature_defs)
	{
		$sql = array();	// each element will hold a piece of the query

		$info = $this->GetInfoSchema($mstable);

		/*
		 * Prepare SELECT expression. Table 'y' is associated with
		 * $mstable while the feature tables are labeled v<N>, where
		 * N is an index starting from 0. Note that the v<N> prefix
		 * is not applied to a feature that is stored directly in 'y'
		 * since such a feature can be selected right away.
		 * Below, the feature table is denoted by $a[0]; the column
		 * name used in the feature table is denoted by $a[1].
		 */
		array_push($sql, 'SELECT y.id,y.model_id');
		$i = 0;
		foreach ($feature_defs as $k => $a) {
			if (array_key_exists($a[1].'_id', $info)) {
				// feature in separate table
				$s = sprintf(",v%d.%s", $i++, $a[1]);
			} else if (array_key_exists($a[1].'_sqn', $info)) {
				// feature in junction table
				$s = sprintf(",%s", $a[1].'_sqn');
			} else {
				// feature in 'y'
				$s = sprintf(",%s", $a[1]);
			}
			array_push($sql, $s);
		}
		/*
		 * Prepare FROM statement. Abbreviate table to 'y'.
		 */
		array_push($sql, 'FROM '.$mstable.' y');
		/*
		 * Add JOIN statements where strings are substituted for IDs.
		 * Note the "_id" postfix on the device type column name, but
		 * only for the features stored separately from 'y' (i.e.,
		 * not in 'y').
		 */
		$i = 0;
		foreach ($feature_defs as $k => $a) {
			if (array_key_exists($a[1].'_id', $info)) {
				$s = sprintf("LEFT JOIN %s v%d ON v%d.id=y.%s_id",
					$a[0], $i, $i, $a[1]);
				array_push($sql, $s);
				$i++;
			}
		}

		return (implode(' ', $sql));
	}

	protected $mHost;	//!< Hostname of database server
	protected $mUser;	//!< User accessing database
	protected $mPass;	//!< Password of user (unencrypted)
	protected $mDatabase;	//!< Name of database to access
	protected $mLink;	//!< Database link identifier
}
?>
