<?php
/*!
 *  @file
 *
 *  @brief	implementation of SQLite3 database access
 *
 *  @date	September 8, 2015
 */

/*!
 *  Implementation of SQLite3 database access for HWS. This class defines
 *  functions, properties, and queries which are specific for the SQLite3
 *  database implementation.
 */
class Sqlite3Database extends SqlDatabase
{
	function __construct($host, $user, $pass)
	{
		parent::__construct($host, $user, $pass);
	}

	function __destruct()
	{
		$this->Close();

		parent::__destruct();
	}

	public function GetDbInfo()
	{
		$db_info = array();

		$db_info['implementation'] = self::DB_IMPL;
		$db_info['host'] = $this->mHost;

		if ($this->mLink) {
			$db_info['state'] = true;
			$db_info['version'] =
				$this->mLink->version()['versionString'];
		} else {
			$db_info['state'] = false;
			$db_info['version'] = '-';
		}

		// for SQLite the uptime is not applicable:
		$db_info['uptime'] = 'N/A';

		return $db_info;
	}

	/*!
	 *  Create SQLite3 database on disk.
	 *
	 *  @param[in] $db      Name of database. This is the full pathname of
	 *                      the database.
	 *  @param[in] $options Database-specific options (optional).
	 *
	 *  @returns
	 *  @c true in case the given database was created; @c false is
	 *  returned in case an error occurred. Note that it's also considered
	 *  an error if the database already exists.
	 */
	public function CreateDatabase($db, $options = '')
	{
		$db_file = trim($db);

		if ($db_file == false) {
			return false;
		}

		if (!file_exists($db_file)) {
			// create the database, but no more than that:
			$s3db = new SQLite3($db_file);
			if ($s3db) {
				$s3db->close();
			}
			return is_file($db_file); // database file created?
		} else {
			return false;		// database already exists
		}
	}

	public function Connect($db)
	{
		// since there's no DBMS to connect to, we're OK
		// when no database was specified:
		if (trim($db) == false) {
			return true;
		}

		// try to open the database R/W:
		try {
			$this->mLink = new SQLite3($db, SQLITE3_OPEN_READWRITE);
			if ($this->mLink) {
				$this->mDatabase = basename($db);
				$ret = true;
			}
		} catch (Exception $e) {
			$ret = false;
		}
		return ($ret);
	}

	public function Close()
	{
		if ($this->mLink) {
			$this->mLink->close();
		}
	}

	public function Insert($table, $fields)
	{
		$fld_values = array();
		$id = 0;

		if (!$this->mLink) {
			return false;
		}

		// format query:
		$cols = implode(',', array_keys($fields));
		$sql  = 'INSERT INTO ' . $table . ' (' . $cols . ') VALUES(';
		foreach ($fields as $k => $val) {
			$parm = $this->FmtSqlParameter($table, $k, $val);
			if ($parm === false) {
				fatal($this->FmtLastError('Insert',
							print_r($parm)));
			}

			array_push($fld_values, $parm);
		}
		$sql .= implode(',', $fld_values) . ')'; // closing parenthesis for VALUES

		$ret = $this->Query($sql);
		if ($ret) {
			$id = $this->mLink->lastInsertRowId();
		} else {
			fatal($this->FmtLastError('Insert', $sql));
		}

		return ($id > 0 ? $id : false);
	}

	public function Update($table, $id, $fields)
	{
		$fld_set = array();

		if (!$this->mLink) {
			return false;
		}

		$sql = 'UPDATE ' . $table . ' SET ';

		// format list of columns to set:
		foreach ($fields as $k => $val) {
			$parm = $this->FmtSqlParameter($table, $k, $val);
			if ($parm === false) {
				fatal($this->FmtLastError('Update',
							print_r($parm)));
			}
			array_push($fld_set, $k.'='.$parm);
		}
		$sql .= implode(',', $fld_set);

		$sql .= ' WHERE id=' . $id;

		$ret = $this->Query($sql);
		if ($ret) {
			return true;
		} else {
			fatal($this->FmtLastError('Update', $sql));
		}
	}

	public function EscapeString($str)
	{
		return $this->mLink->escapeString($str);
	}

	public function FmtSqlParameter($table, $column, $value)
	{
		if (is_null($value))    return 'NULL';
		// check null first as it's not considered scalar

		if (!is_scalar($value)) return false;

		$info = $this->GetInfoSchema($table);
		switch ($info[$column]['type']) {
		case 'bool':
		case 'int':
			return (int)$value;
		case 'float':
			return (float)$value;
		case 'string':
			return '\''.$this->EscapeString($value).'\'';
		default:
			return $value;
		}
	}

	public function FmtLastError($context, $data)
	{
		$err  = '<h4>Error</h4>';
		$err .= $this->mLink->lastErrorMsg();
		$err .= ' ('.$this->mLink->lastErrorCode().')';
		$err .= '<h4>Database</h4>';
		$err .= $this->mDatabase;
		$err .= '<h4>Backend</h4>';
		$err .= self::DB_IMPL;
		$err .= '<h4>Context</h4>';
		$err .= empty($context) ? 'N/A' : $context;
		$err .= '<h4>Context Data</h4>';
		$err .= empty($data) ? 'N/A' : filter_string($data);

		return $err;
	}

	public function GetInfoSchema($table)
	{
		$info = $this->GetInfoSchemaRaw($table);

		foreach ($info as &$col_info) {
			switch (strtolower($col_info['type'])) {
			case 'boolean':
				$col_info['type'] = 'bool'; break;
			case 'int':
			case 'integer':
			case 'tinyint':
			case 'smallint':
			case 'mediumint':
			case 'bigint':
			case 'int2':
			case 'int8':
				$col_info['type'] = 'int'; break;
			case 'float':
			case 'double':
			case 'real':
			case 'decimal':
			case 'numeric':
				$col_info['type'] = 'float'; break;
			default:
				$col_info['type'] = 'string'; break;
			}
		}
		unset($col_info);

		return $info;
	}

	/*!
	 *  \name Physical Devices
	 *  @{
	 */

	public function GetNumberOfPhysicalDevices($gid, $tid, $msid)
	{
		// format device filter statements, if necessary:
		$filters = array();
		if (!empty($gid)) {
			array_push($filters, 'hwgroup_id='.$gid);
		}
		if (!empty($tid)) {
			array_push($filters, 'device_type_id='.$tid);
		}
		if (!empty($msid)) {
			array_push($filters, 'model_spec_id='.$msid);
		}

		// query to filter physical devices:
		$sql = 'SELECT id,count(*) as count FROM physical';
		if (count($filters) > 0) {
			$sql .= ' WHERE ' . implode(' AND ', $filters);
		}
		$ret = $this->Query($sql);
		if ($ret) {
			return $ret[0]['count'];
		} else {
			return false;
		}
	}

	public function GetPhysicalDeviceId($gid, $tid, $sn)
	{
		$sql  = 'SELECT id FROM physical';
		$sql .= ' WHERE hwgroup_id=:gid';
		$sql .= '   AND device_type_id=:tid';
		$sql .= '   AND serial_number=:sn';
		$stm = $this->mLink->prepare($sql);
		$stm->bindValue(':gid', $gid, SQLITE3_INTEGER);
		$stm->bindValue(':tid', $tid, SQLITE3_INTEGER);
		$stm->bindValue(':sn',  $sn,  SQLITE3_TEXT);
		$result = $stm->execute();
		if ($result) {
			$rows = array();
			while ($r = $result->fetchArray(SQLITE3_ASSOC)) {
				array_push($rows, $r);
			}

			$nmatches = count($rows);
			if ($nmatches > 1) {		// multiple matches
				$ret_val = -$nmatches;
			} else if ($nmatches == 1) {	// one match
				$ret_val = $rows[0]['id'];
			} else {			// no match
				$ret_val = 0;
			}
		} else {
			$ret_val = 0;
		}
		return ($ret_val);
	}

        /*!
	 *  end Physical Devices
	 *  @}
	 */

	protected function Query($sql)
	{
		$rows = array();

		if (!$this->mLink) {
			return false;
		}

		/*
		 * Return value of query() method doesn't seem to be correct
		 * (PHP 5.4.41). TRUE should be returned if the SQL query
		 * returns no value, but instead an empty SQLite3Result object
		 * is returned. The caller may interpret this empty object as
		 * FALSE, thus a failure. So, if the query succeeds but yields
		 * no results, we'll force the return value of this function
		 * to TRUE.
		 */
		$result = $this->mLink->query($sql);
		if ($result === false) {
			fatal($this->FmtLastError('Query', $sql));
		}

		if ($result === true) {
			return $result;
		} else {
			// retrieve results:
			if ($result->numColumns() > 0) {
				while ($r = $result->fetchArray(SQLITE3_ASSOC)) {
					array_push($rows, $r);
				}
			} else {
				$rows = true;
			}
			$result->finalize();

			return $rows;
		}
	}

	/*!
	 *  Return data type of column. This is a helper function which
	 *  filters the data type length that may be part of the data type
	 *  specification. The length is assumed to be given in parenthesis.
	 *
	 *  @param[in] $str Column definition string.
	 *
	 *  @returns
	 *  Data type string.
	 */
	private function GetCharType($str)
	{
		$regex = '/\([0-9]+\)/';

		return preg_replace($regex, '//', $str);
	}

	/*!
	 *  Determine length of a @c CHAR type column. This is a helper
	 *  function which tries to extract the length part of a
	 *  character-based table column. Such a column has the substring
	 *  @c CHAR in its name and the maximum length in parenthesis.
	 *
	 *  @param[in] $str Column definition string.
	 *
	 *  @returns
	 *  Maximum length of @c CHAR column or 0 if undetermined.
	 */
	private function GetCharTypeLength($str)
	{
		$regex = '/^[A-Z]*CHAR[A-Z]*\(([0-9]+)\)$/i';

		if (preg_match($regex, $str, $match)) {
			return $match[1];
		} else {
			return 0;
		}
	}

	/*!
	 *  Retrieve untranslated information schema of table. This function
	 *  is the same as GetInfoSchema() except the table information is
	 *  not translated.
	 *
	 *  @sa
	 *  GetInfoSchema()
	 */
	private function GetInfoSchemaRaw($table)
	{
		$info = array();

		$sql = 'PRAGMA table_info('.$table.')';
		$ret = $this->Query($sql);
		if ($ret) {
			foreach ($ret as $row) {
				$name = $row['name'];
				$type = $this->GetCharType($row['type']);
				$len  = $this->GetCharTypeLength($row['type']);

				// prepare results in the proper format:
				$info[$name] = array(
						'type' => $type,
						'max_length' => $len,
						'default' => $row['dflt_value']
					       );
			}
		}
		return $info;
	}

	const DB_IMPL = 'SQLite3';
}
?>
