<?php
/*!
 *  @file
 *
 *  @brief	functions related to authentication
 *
 *  @date	September 3, 2015
 */

/* Decrypt a given (encrypted) password.
 */
function decrypt_password($crypt_password)
{
	$key = get_hws_key();
	if (empty($key)) fatal(E_NOKEY);	/* Need key to decode. */
	
	$dcrypt_passwd = @mcrypt_decrypt(MCRYPT_TRIPLEDES, $key,
					 $crypt_password, MCRYPT_MODE_ECB);
	
	/* In case of a short password, the decrypted password is padded
	 * with NULLs. These need to be trimmed. */
	$dcrypt_passwd = rtrim($dcrypt_passwd, "\0");
	
	return ($dcrypt_passwd);
}

/* Encrypt the given (cleartext) password.
 */
function encrypt_password($password)
{
	$key = get_hws_key();
	if (empty($key)) fatal(E_NOKEY, true);
	
	$crypt_passwd = @mcrypt_encrypt(MCRYPT_TRIPLEDES, $key, $password,
					MCRYPT_MODE_ECB);
	return ($crypt_passwd);
}

/* Retrieve the site-permissions for a user from the HTTP Authorization
 * variable. Any required data missing leads to read-only access. A password
 * mispatch leads to empty permissions (i.e. '-').
 */
function get_credentials()
{
	global $DB;

	/* Get the Authorization header parameters. */
	$all_hdrs = getallheaders();
	$auth_hdr = $all_hdrs['Authorization'];
	if (!isset($auth_hdr)) return '';

	$nonce = $_SESSION['nonce'];
	if (empty($nonce)) return '';

	/* Get response digest of client (part of Authorization header). */
	$response = get_directive_value($auth_hdr, "response");
	if (empty($response)) return '';

	/* Get username: first try session data, else Authorization header. */
	$user = $_SESSION['user'];
	if (empty($user)) $user = get_directive_value($auth_hdr, "username");
	if (empty($user)) return '?';

	/* Check whether user exists. */
	$uid = $DB->GetUserId($user);
	if ($uid == 0) return '?';
	$row = $DB->GetUserInfo($uid);
	if (empty($row)) return '?';

	/*
	 * Only check for a locked account if supported
	 * by the database (backwards compatibility).
	 */
	if ($row['locked'] != NULL) {	/* not using empty() as  */
					/* 'locked' may be false */
		if ($row['locked'] == true) {
			return '!';
		}
	}

	if (empty($row['password'])) {
		$dcrypt_passwd = '';
	} else {
		$dcrypt_passwd = decrypt_password(base64_decode($row['password']));
	}

	/* Create request-digest. */
	$a1 = $user . ":" . HWS_REALM . ":" . $dcrypt_passwd;
	$a2 = $_SERVER['REQUEST_METHOD'] . ":" . $_SERVER['REQUEST_URI'];
	$rd = md5(md5($a1) . ":" . $nonce . ":" . md5($a2));

	/* Compare given reponse (client) with calculated digest (server). */
	if (strcasecmp($response, $rd) == 0) {
		return $row['permissions'];
	} else {
		return '-';
	}
}

/*!
 *  Check requested permissions for user. The credentials of the current user
 *  are compared to a given set of access permissions. This is translated to a
 *  boolean result code indicating whether or not the user satisfies the
 *  requested permissions. The credentials of the user must at least match the
 *  requested permissions in order for this function to return a positive
 *  result.
 *
 *  @param[in] $req_perm List of requested permissions.
 *
 *  @returns
 *  \c true if the credentials of the user satisfy the requested permissions,
 *  otherwise \c false.
 */
function has_permissions($req_perm)
{
	$user_perm = get_credentials();

	foreach (str_split($req_perm) as $p) {
		if (strpos($user_perm, $p) === false) {
			return (false);
		}
	}
	return (true);
}

/* Return the value of the given directive from the HTTP Authorization
 * variable.
 */
function get_directive_value($auth_hdr, $directive)
{
	/* Make array of the parameters of $auth_hdr (comma separated). */
	$auth_r = preg_split("/,/", $auth_hdr);
	if (empty($auth_r) || empty($directive)) return (null);
	
	/* Find given directive and save its value (separated by "="). */
	$pattern = "/\s*" . $directive. "/"; /* 0 or more whitespace prefix */
	foreach ($auth_r as $s) {
		if (preg_match($pattern, $s)) {
			/* Directive found: assign its value to $value. */
			list($dummy, $value) = explode("=", $s, 2);
			$value = trim($value, '"');
			if (!empty($value)) break;
		}
	}
	return ($value);
}

/* Read the secret key for encrypting/decrypting user passwords.
 */
function get_hws_key()
{
	$f = @fopen(HWS_KEY, "r");
	if (!$f) return (null);

	$k = fgets($f);
	if (!$k) $k = "";	/* Change from false to null. */
	
	fclose($f);
	return ($k);
}

function get_nonce_key()
{
	$f = @fopen(NONCE_KEY, "r");
	if (!$f) return (null);

	$k = fgets($f);
	if (!$k) $k = "";	/* Change from false to null. */
	
	fclose($f);
	return ($k);
}
?>
