<?php
/*!
 *  @file
 *
 *  @brief	functions related to data retrieval and manipulation
 *
 *  @date	2005
 */

/*!
 *  Convert harmfull tokens in string. Basic HTML tags in the given string are
 *  and convert to HTML entities. These include:
 *
 *  - <code>&</code>  (ampersand)
 *  - <code>"</code>  (double quote)
 *  - <code>'</code>  (single quote)
 *  - <code>\<</code> (lower than)
 *  - <code>\></code> (greater than)
 *
 *  @param[in] $str String to be converted.
 *
 *  @returns
 *  The converted string.
 */
function filter_string($str)
{
	$s = $str;
	
	/*
	 * In case magic_quotes_gpc is set, escapes (\) were added to the
	 * string. These escapes are removed before filtering.
	 */
	if (get_magic_quotes_gpc()) $s = stripslashes($s);
	
	return (htmlentities($s, ENT_QUOTES));
}

/*!
 *  Check for non-empty string. The given string is only valid if it's not
 *  empty or contains one or more non-whitespace characters. If it does,
 *  the script is aborted with a fatal error.
 *
 *  @param[in] $str String to be checked. Leading and/or trailing
 *                  whitespace on the string is ignored.
 *
 *  @returns
 *  If the string is considered valid, it is returned. Any leading or
 *  trailing whitespace is removed before the string is returned. If the
 *  string is not valid, the script aborts with a fatal error (\c E_NOPARM).
 */
function validate_string($str)
{
	$s = trim($str);

	if (!empty($s)) return $s;

	fatal(E_NOPARM);
}

/*!
 *  Check if string is an integer number.
 *
 *  @param[in] $str String to be checked. Leading and/or trailing
 *                  whitespace on the string is ignored.
 *
 *  @returns
 *  \c true if the string is an integer, otherwise \c false.
 */
function is_decimal($str) // is_int() / is_integer() already defined
{
	$s = trim($str);

	if (!empty($s) && ctype_digit($s)) return (true);

	return (false);
}

/*!
 *  Enforce string is an integer number. If the given string is not an
 *  integer, the function will abort the script with a fatal error.
 *
 *  @param[in] $str String to be checked. Leading and/or trailing
 *                  whitespace on the string is ignored.
 *
 *  @returns
 *  The integer number is returned if it's indeed an integer. The returned
 *  string is also stripped from any whitespace. If the string is not an
 *  integer, then a fatal error occurs (\c E_NONAN).
 */
function validate_integer($str)
{
	if (is_decimal($str)) return (trim($str));

	fatal(E_NONAN);
}

/*!
 *  Convert string to boolean value. Common representations of a
 *  boolean value are recognized. Textual representations are not
 *  case-sensitive.
 *
 *  - true / false
 *  - yes / no
 *  - 0 / 1
 *  - on / off
 *  - enabled / disabled
 *
 *  @param[in] $str String to be converted. Leading and/or trailing
 *                  whitespace on the string is ignored.
 *
 *  @returns
 *  \c true or \c false depending value of given string. If the string does
 *  not represent a boolean value, then a NULL string is returned.
 */
function to_boolean($str)
{
	$val = trim(strtolower($str));

	switch ($val) {
	case 'true':
	case 'yes':
	case '1':
	case 'on':
	case 'enabled':
		return true;
	case 'false':
	case 'no':
	case '0':
	case 'off':
	case 'disabled':
		return false;
	default:
		return NULL;
	}
}

/*!
 *  Check if string is a boolean value.
 *
 *  @param[in] $str String to be checked. Leading and/or trailing
 *                  whitespace on the string is ignored.
 *
 *  @returns
 *  \c true if the string is a boolean, otherwise \c false. Note that the
 *  function indicates whether or not the given string @e represents a boolean;
 *  the returned value is @e not the evaluated value of the string.
 */
function is_boolean($str)
{
	if (!is_null(to_boolean($str))) return true;

	return false;
}

/*!
 *  Enforce string is a boolean value.
 *
 *  If the given string is a boolean value then its value will be translated
 *  to the corresponding boolean data type of PHP (i.e., \c true or \c false).
 *
 *  @param[in] $str String to be checked. Leading and/or trailing whitespace
 *                  on the string is ignored.
 *
 *  @returns
 *  \c true or \c false depending value of given string. If the string does
 *  not represent a boolean value, then a fatal error occurs (\c E_NONAN).
 */
function validate_boolean($str)
{
	if (is_boolean($str)) return to_boolean($str);
	
	fatal(E_NONAN);
}

/*!
 *  Check if string is a proper filename. The function uses a limitted
 *  range of characters that are allowed in a filename. This will probably
 *  differ from the underlying filesystem, so bear in mind this function
 *  may reject a filename although it would be valid on the filesystem
 *  being used.
 *
 *  For this function a filename is valid if it starts with a letter
 *  (upper- or lowercase), digit, or underscore. Any subsequent character
 *  may be a letter (upper- or lowercase), digit, underscore, minus, or
 *  period.
 *
 *  @param[in] $str String to be checked. Leading and/or trailing
 *                  whitespace on the string is ignored.
 *
 *  @returns
 *  If the string represents a valid filename, the string is returned.
 *  Otherwise a fatal error is generated (\c E_FILENAME).
 */
function validate_filename($str)
{
	$patt = '/^[A-Za-z0-9_][-.A-Za-z0-9_]*$/';
	$s = trim($str);

	if (!empty($s) && preg_match($patt, $s)==1) return $s;

	fatal(E_FILENAME);
}

/*!
 *  Express feature value in given unit. The unit specified in the feature
 *  value is looked up in the unit mapping table. The scaling factor
 *  associated with the unit is used to express the value in the given unit.
 *  The layout of the unit mapping table is defined elsewhere in the
 *  documentation.
 *
 *  If the unit in the feature value is omitted (which doesn't really make
 *  sense), then no conversion will be done and the value is returned as-is.
 *  Units are case-insensitive.
 *  
 *  @param[in] $value      Feature value with associated unit.
 *  @param[in] $unit       Convert value to this unit. The specified unit must
 *                         correspond to one of the units in the unit mapping
 *                         table.
 *  @param[in] $unit_table Name of database table holding the unit mappings.
 *
 *  @returns
 *  Converted feature value if successful. The unit is no longer part of the
 *  return value, so only the value is returned. @c false is returned in case
 *  the conversion failed.
 *
 *  @sa
 *  File: <code>doc/dtd.txt</code>.
 */
function convert_value($value, $unit, $unit_table)
{
	global $DB;
	$to_unit = trim($unit);

	if (empty($unit_table) || empty($to_unit)) {
		return false;
	}

	// get entire mapping table:
	$map = $DB->FetchByValue($unit_table);
	if (!$map) {
		return false;
	}

	// match number and unit in feature value string:
	$pattern = '/^[[:space:]]*([0-9.]+)[[:space:]]*([A-Za-z]*)[[:space:]]*$/';
	$ret = preg_match($pattern, $value, $matches);
	if ($ret == 1) {
		// determine factor of feature value and unit parameter:
		foreach ($map as $v) {
			if (strcasecmp($matches[2], $v['unit']) == 0) {
				$from_factor = $v['factor'];
			}
			if (strcasecmp($to_unit, $v['unit']) == 0) {
				$to_factor = $v['factor'];
			}
		}
		if (!empty($from_factor) && !empty($to_factor)) {
			$to_value = $matches[1] * $from_factor / (float)$to_factor;
		} else {
			$to_value = $matches[1];
		}
		// return as "string" type to avoid
		// conversion from float to int
		return (string)$to_value;
	} else {
		return false;
	}
}

/*!
 *  Read model features from variable list. The model features and their
 *  value are stored in an associative array and returned to the caller.
 *  Each feature in the array is stored as follows:
 *  <pre><code>
 *  \<feature_identifier\> => (
 *                              'table'  => \<feature_table\>,
 *                              'column' => \<feature_column\>,
 *                              'value'  => \<feature_value\>
 *                            )
 *  </code></pre>
 *
 *  The function scans the variable list for a given set of features. The
 *  features are identified by their name in the DTD (i.e., member
 *  <code>feature_column</code>) although a callback function can be used
 *  if a different naming scheme is being used. Given the correct feature
 *  name, its value can be read from the variable list and stored in the
 *  array structure listed above. The <code>feature_value</code> member
 *  holds an array if there are multiple values set for a given feature.
 *  Otherwise, <code>feature_value</code> is just a string value.
 *
 *  @param[in] $feature_defs Array defining the features to look for. This
 *                           array has the same layout as the "features"
 *                           array of a DTD.
 *  @param[in] $var_list     List of variables presumably holding the model
 *                           features and their value. The list must be an
 *                           associative array.
 *  @param[in] $fn_xlate     Optional callback function for translating or
 *                           formatting the correct feature names. This is
 *                           useful in case some naming convention has been
 *                           applied to the feature names and you need to
 *                           relate those names to the original feature
 *                           names. The returned array (mentioned above)
 *                           always uses the original feature names. If no
 *                           callback function is specified then the
 *                           original feature names are used unmodified. The
 *                           callback function takes a copy of $var_list as
 *                           its first argument, the original feature name
 *                           as its second argument, and the data type of
 *                           the feature as its third argument.
 *
 *  @returns
 *  Associative array holding the model features found in $var_list or an
 *  empty array if nothing found. The function may also abort the script if
 *  a required feature is missing.
 *
 *  @sa
 *  File: <code>doc/dtd.txt</code>.
 */
function extract_features($feature_defs, $var_list, $fn_xlate = '')
{
	$feats = array();

	foreach ($feature_defs as $k => $s) {
		// $s[0]: feature table
		// $s[1]: feature column name
		// $s[3]: datatype of column
		// $s[5]: input required (flag)

		// determine feature name(s):
		if (empty($fn_xlate)) {
			$ftname = $s[1];
		} else {
			$ftname = $fn_xlate($var_list, $s[1], $s[3]);
		}

		// read, check and store value(s):
		if (is_array($ftname)) {
			$val = array();
			foreach ($ftname as $v) {
				if ($s[5]) $val[] = validate_string($var_list[$v]);
				else $val[] = $var_list[$v];
			}
		} else {
			$val = '';
			if ($s[5]) $val = validate_string($var_list[$ftname]);
			else $val = $var_list[$ftname];
		}
		$feats[$k] = array('table'  => $s[0],
				   'column' => $s[1],
				   'value'  => $val);
	}
	return $feats;
}

/*!
 *  Read model features from FORM. This is a wrapper function for
 *  extract_features() but tailored for features from a FORM.
 *
 *  @sa
 *  extract_features()
 */
function extract_features_form($feature_defs, $var_list)
{
	return extract_features($feature_defs,
				$var_list,
				'cb_fmt_ftname_form');
}

/*!
 *  Callback function translating feature names for FORMs.
 *
 *  @param[in] $var_list List of variables holding model
 *                       features and their value.
 *  @param[in] $ftname   Original feature name.
 *  @param[in] $type     Data type of feature.
 *
 *  @returns
 *  Feature name(s) in $var_list corresponding to $ftname.
 *  If there are multiple feature names, they are returned
 *  in an array.
 */
function cb_fmt_ftname_form($var_list, $ftname, $type)
{
	switch ($type) {
	case TEXT:
		$xname = $ftname . '_text';
		break;
	case MENU:
		$xname = $ftname . '_menu';
		if (strcmp($var_list[$xname], 'other') == 0) {
			// use text input if "other" selected:
			$xname = $ftname . '_text';
		}
		break;
	case RBTN:
		$xname = $ftname . '_rbtn';
		break;
	case CBOX:
		# filter checkbox ("cbox") parameters from list:
		$xname = preg_grep('/' . $ftname . '_cbox_[0-9]+/',
				   array_keys($var_list));
		break;
	default:
		fatal(E_DATATYPE);
	}
	return $xname;
}

/*!
 *  Ensure feature is available in database table. The purpose of this function
 *  is to determine the ID for a list of device features. This means the table
 *  of a given feature is searched for a certain entry or value. If the value is
 *  found, its ID is stored on an ID-list. This is repeated for all the device
 *  features on the given list. In case the value could not be found, it is
 *  added to the table and its ID is stored on the ID-list. A value cannot be
 *  empty (null); the value will not be added to the table nor will an ID be
 *  added to the ID-list. The corresponding feature is effectively ignored.
 *
 *  It's also possible a feature is not stored by ID, but directly in a
 *  model-specific table. In this case the feature and its value are also
 *  stored on the ID-list although the value does not represent an ID. This
 *  way the list of features will remain complete. Again, an empty value is
 *  not added to the ID-list.
 *
 *  @param[in] $mstable  Name of model-specific table. This is the table where
 *                       all the feature IDs are saved to identify a
 *                       particular model.
 *  @param[in] $features Array holding a list of table names, and
 *                       feature/value pairs. Refer to extract_features() for
 *                       the exact layout of the array.
 *
 *  @returns
 *  Associative array holding feature/ID pairs and/or feature/value pairs.
 *  Keys in the array are named after a feature, but note that the key of a
 *  feature/ID pair is post-fixed with the literal "_id". The key name of a
 *  feature/value pair is not altered. So, the returned array consists of
 *  elements of the following form:
 *  <pre><code>
 *  (
 *    {\<feature\>_id => \<id\>} || {\<feature\> => \<value\>}
 *    ...
 *  )
 *  </code></pre>
 */
function resolve_feature_ids($mstable, $features)
{
	global $DB;
	$ids = array();

	$info = $DB->GetInfoSchema($mstable);

	foreach ($features as $s_r) {
		$colname    = $s_r['column'];
		$colname_id = $colname . '_id';

		/*
		 * In case the feature is saved by ID in the model-specific
		 * table then we lookup the ID in the corresponding feature
		 * table or add it to that table, if necessary. Otherwise,
		 * the feature is directly saved in the model-specific table
		 * and we preserve the value of the feature.
		 */
		if (array_key_exists($colname_id, $info)) {
			$arr = array($s_r['column'] => $s_r['value']);

			$ret = $DB->FetchByValue($s_r['table'], $arr);
			if ($ret) {
				$id = $ret[0]['id'];
			} else {
				if (!empty($s_r['value'])) {
					$id = $DB->Insert($s_r['table'], $arr);
				} else {
					$id = 0;
				}
			}

			// prevent adding empty or failed cases:
			if ($id > 0) {
				$ids[$colname_id] = $id;
			}
		} else {
			// prevent adding empty values:
			if (!empty($s_r['value'])) {
				$ids[$colname] = $s_r['value'];
			}
		}
	}
	return $ids;
}

/*!
 *  Ensure feature is available in junction table. This function resembles
 *  resolve_feature_ids() except it operates on features stored in a junction
 *  table. The junction table allows multiple selections of a given feature
 *  (e.g., I/O busses) to be stored under a common identifier (here, the
 *  identifier is referred to as a <em>sequence number</em>).
 *
 *  Given a list of features, the function only considers those features that
 *  must be stored in a junction table. If a selection of values of a certain
 *  feature is already present in the junction table, then its sequence number
 *  is stored on a list. Otherwise, the selection is added to the junction
 *  table under a new sequence number. The new sequence number is also stored
 *  on the aforementioned list. The result that is returned consists only of
 *  sequence numbers.
 *
 *  Note that it's allowed to have an empty selection for a feature. In this
 *  case no sequence number will be recorded in the junction table.
 *
 *  @param[in] $mstable  Name of model-specific table. This is the table where
 *                       all the feature IDs are saved to identify a
 *                       particular model.
 *  @param[in] $features Array holding a list of table names, and
 *                       feature/value pairs. Refer to extract_features() for
 *                       the exact layout of the array.
 *
 *  @returns
 *  Associative array holding feature/sequence number pairs. Keys in the array
 *  are named after a feature, but note that the key is post-fixed with the
 *  literal "_sqn" (junction table sequence number). The key name of a
 *  feature/sequence number pair is not altered. The returned array consists of
 *  elements of the following form:
 *  <pre><code>
 *  (
 *    {\<feature\>_sqn => \<seq_number\>}
 *    ...
 *  )
 *  </code></pre>
 */
function resolve_jtable_sqn($mstable, $features)
{
	global $DB;
	$sqn = array();

	$info = $DB->GetInfoSchema($mstable);

	foreach ($features as $s_r) {
		$colname     = $s_r['column'];
		$colname_sqn = $colname . '_sqn';

		if (array_key_exists($colname_sqn, $info)) {
			$values = $s_r['value']; // list of feature values
			if (!empty($values)) {
				$ret = $DB->FetchByValue('feature_tables',
						array('name' => $s_r['table']));
				$tbl_id = $ret[0]['id'];

				$num = jt_get_seq_number($s_r['table'],
						      $colname,
						      $values);
				if ($num == 0) {
					$num = jt_get_next_seq_number();

					$ret = $DB->FetchByMultiValue(
							$s_r['table'],
							$colname,
							$values);
					// insert features one-by-one in jtable:
					foreach ($ret as $v) {
						$arr = array(
						  'seq_number'     => $num,
						  'feature_tbl_id' => $tbl_id,
						  'feature_id'     => $v['id']);
						$DB->Insert('feature_jt', $arr);
					}
				}

				// prevent adding empty or failed cases:
				if ($num > 0) {
					$sqn[$colname_sqn] = $num;
				}
			}
		}
	}
	return $sqn;
}

/*!
 *  Return next sequence number. The value returned by this function is the
 *  next unused sequence number in the junction table <code>feature_jt</code>.
 *  A sequence number is a positive integer, starting from 1.
 *
 *  @returns
 *  Next sequence number.
 */
function jt_get_next_seq_number()
{
	global $DB;

	$ret = $DB->Fetch('feature_jt', array('seq_number'), 'seq_number');
	if (empty($ret)) {
		$sqn = 1;
	} else {
		$sqn = end($ret)['seq_number'] + 1;
	}
	return $sqn;
}

/*!
 *  Determine sequence number for selection of feature values. This function
 *  searches the junction table (<code>feature_jt</code>) for the given
 *  selection of feature values. If found, the associated sequence number is
 *  returned.
 *
 *  @param[in] $table  Name of feature table.
 *  @param[in] $column Column in feature table holding the feature values.
 *  @param[in] $values Array containing the feature values to be searched for.
 *                     These values correspond to the values stored in $column
 *                     of $table.
 *
 *  @returns
 *  Sequence number of matching selection or 0 if not found.
 */
function jt_get_seq_number($table, $column, $values)
{
	global $DB;
	$ids = array();
	$my_ids = array();
	$sqn = 0;

	$ret = $DB->FetchByValue('feature_tables', array('name' => $table));
	$tbl_id = $ret[0]['id'];
	if (!empty($tbl_id)) {
		// get IDs of features:
		$ret = $DB->FetchByMultiValue($table, $column, $values);
		foreach ($ret as $v) {
			$my_ids[] = $v['id'];
		}
		if (!empty($my_ids)) {
			// get sequence numbers associated with this table:
			$arr = array('feature_tbl_id' => $tbl_id);
			$ret = $DB->FetchByValue('feature_jt', $arr);
			// group feature IDs by sequence number:
			foreach ($ret as $v) {
				$tmp = $v['seq_number'];
				$ids[$tmp][] = $v['feature_id'];
			}
			// find our set of feature IDs:
			foreach ($ids as $k => $v) {
				if (count($v) == count($my_ids) &&
				    !array_diff($v, $my_ids))
				{
					// found it, save seq. number:
					$sqn = $k;
					break;
				}
			}
			
		}
	}
	return $sqn;
}

/*!
 *  Filter URL parameter. Format a parameter list for the given URL, but omit
 *  a specified parameter and its value.
 *
 *  @param[in] $url  URL to be filtered.
 *  @param[in] $parm Parameter to be filtered from $url.
 *
 *  @returns
 *  URL without the parameter given by $parm, if it existed. Otherwise, the
 *  URL is returned unchanged.
 */
function url_strip_parm($url, $parm)
{
	$parm_list = '';
	
	// turn URL into array (elements split at '&'):
	$parms = preg_split('/&/', $url);

	// paste parameters together, but omit $parm:
	foreach ($parms as $s) {
		if (strpos($s, $parm) === false && !empty($s))
			$parm_list .= $s . '&';
	}

	// remove trailing '&':
	$parm_list = rtrim($parm_list, '&');
	
	return ($parm_list);
}
?>
