#!/usr/local/bin/php
<?php
/*!
 *  @file
 *
 *  @brief	add empty device type definition to HWS site
 *
 *  <p>
 *  <code>
 *  usage: create_new_dtd.php [OPTIONS] -- \<DTD\>
 *  </code>
 *  </p>
 *
 *  @returns
 *  Exit code 0 is returned if the new DTD was created successfully,
 *  otherwise code 1.
 *
 *  @date	August 24, 2015
 *
 *  @todo
 *  Eliminate the use of the "--" marker. Preferably there should be no
 *  marker separating command line options and the DTD name.
 */

define("VERSION", "0.1");

//-------------------------------- functions -----------------------------------

function version()
{
	global $my_basename;

	fprintf(STDOUT, "%s v%s\n", $my_basename, VERSION);

	exit(1);
}

function usage()
{
	global $my_basename;

	fprintf(STDOUT, "
usage: %s [OPTIONS] -- DTD

Create a new device type definition file for the HWS system. The name of the
device type definition file is given by the DTD parameter. This parameter is
merely the name of the DTD file; the exact location where DTDs are stored is
determined by the site's configuration file.
This script only creates a template DTD file, so afterwards the site maintainer
has to fill out the specifics for the DTD. By default, the features
\"manufacture\", and \"model\" are added to the DTD.

OPTIONS

  -d DOCROOT     use document root DOCROOT of HWS system (default: .)
  -f             force operation (e.g., overwrite existing DTD file)
  -t TABLE       name of database table for DTD
  -C             do not add identifiers for common features to DTD
  -F f1[,f2]...  list of feature identifiers to add; if omitted one empty
                 feature entry is automatically added
  -h             show this help and exit
  -V             show version and exit
", $my_basename);

	exit(1);
}

// extract DTD name from argument list
function get_dtd_name($arg_list)
{
	$dtd = '';

	// DTD argument follows '--' marker:
	$pos = array_search('--', $arg_list);
	if ($pos && $pos+1 <= count($arg_list)) {
		$dtd = trim($arg_list[$pos+1]);
	}
	return $dtd;
}

//----------------------------- start of script --------------------------------

$short_opts  = 'd:ft:CF:hV';
$my_basename = basename($argv[0]);

$docroot  = '.';
$force    = false;
$table    = '';
$commons  = true;
$features = array();
$dtd      = '';

// handle command line options:
$options = getopt($short_opts);
if ($options) {
	foreach ($options as $opt => $val) {
		switch ($opt) {
		case 'd': $docroot = $val; break;
		case 'f': $force = true; break;
		case 't': $table = trim($val); break;
		case 'C': $commons = false; break;
		case 'F': $features = explode(',', $val); break;
		case 'h': usage(); break;
		case 'V': version(); break;
		}
	}
} else {
	/*
	 * Poor error handling on getopt()'s part. As long as there's one
	 * valid option, getopt() returns "OK", but it ignores invalid
	 * options. So, the message below will not catch all syntax errors.
	 */
	fprintf(STDERR, "%s: syntax error\n", $my_basename);
	usage();
}

$dtd = get_dtd_name($argv);
if (empty($dtd)) {
	fprintf(STDERR, "%s: missing DTD name\n", $my_basename);
	usage(1);
}

// work from within document root:
if (is_dir($docroot)) {
	chdir($docroot) || exit(1);
} else {
	fprintf(STDERR, "%s: %s: directory not found\n", $my_basename,$docroot);
	exit(1);
}

require 'init.php';


/*
 * Create the DTD file.
 */
$dtd_path = DTD_DIR . '/' . $dtd . '.' . DTD_EXT;
if (file_exists($dtd_path) && !$force) {
	fprintf(STDERR, "%s: %s: DTD already exists\n", $my_basename, $dtd);
	exit(1);
}

$fp = fopen($dtd_path, 'w+');
if (!$fp) {
	fprintf(STDERR, "%s: %s: unable to create file\n", $my_basename, $dtd);
	exit(1);
}

fprintf($fp,
"<?php
/*!
 *  @file
 *
 *  @brief	device type definition for %s
 *
 *  @date	%s
 *
 *  @note
 *  This file was generated by %s v%s.
 */

\$DTD = array(
  'table'    => '%s',
  'features' => array(\n",
	$dtd, @date('F d, Y'), $my_basename, VERSION, $table);

if ($commons) {
	fprintf($fp,
"    'manufacture' => array('manufacture',   'manufacture', 'Manufacture', TEXT, false),
    'model'       => array('model',         'model',       'Model',       TEXT, true),\n");
}

// either add entries for requested features or an empty entry:
if (!empty($features)) {
	/*
	 * Save last key. We need this to predict the last iteration of the loop
	 * below, so we can withhold outputting the comma on the last entry.
	 */
	end($features);
	$last_k = key($features);

	foreach ($features as $k => $v) {
		fprintf($fp,
			"    '%s' => array('%s', '%s', '%s', TEXT, false)%s\n",
			$v, $v, $v, ucfirst($v), $k != $last_k ? ',' : '');
	}
} else {
	fprintf($fp, "    '' => array('', '', '', TEXT, false)\n");
}

fprintf($fp,
"  )
);
?>");

fclose($fp);

exit(0);
?>
